/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package cli;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

public class CLIOptions implements ICLIOptions {
	
	@SuppressWarnings("unchecked")
	private Hashtable<String, ICLIOption> options;
	private String[] args;
	private String longPrefix = "--";
	private String shortPrefix = "-";
	private ArrayList<String> unmatched;
	@SuppressWarnings("unchecked")
	private ArrayList<ICLIOption> mandatory;
	private LinkedList<String> optionsKeys;
	private int maxLongLength = 0;
	private int maxTypeLength = 0;
	
	@SuppressWarnings("unchecked")
	public CLIOptions(String[] arguments) {
		options = new Hashtable<String, ICLIOption>();
		args = arguments;
		unmatched = new ArrayList<String>();
		mandatory = new ArrayList<ICLIOption>();
		optionsKeys = new LinkedList<String>();
	}
	
	@SuppressWarnings("unchecked")
	public boolean parse() {
		for(int i = 0; i <  args.length;) {
			ICLIOption option = findMatchingOption(args[i]);
			if(option != null) {
				if(mandatory.contains(option))
					mandatory.remove(option);
				if(option.getSkip() == 1) {
					option.setValue(args[i]);
					i++;
				} else if(i+1 <= args.length) {
					if((i+1 < args.length) && (option.isValidValue(args[i+1]))) {
						option.setValue(args[i+1]);
						i += 2;
					} else { 
						if(option.ignoreInvalid()) {
							option.setValue(args[i]);
							i++;
						} else {
							addUnMatchedParameters(i);
							break;
						}
					}
				}
			} else {
				addUnMatchedParameters(i);
				i++;
				break;
			}
		}
		if(!mandatory.isEmpty()) {
			return false;
		}
		return true;
	}
	
	private void addUnMatchedParameters(int k) {
		for(int i = k; i <  args.length; i++) {
			unmatched.add(args[i]);
		}
	}

	@SuppressWarnings("unchecked")
	private ICLIOption findMatchingOption(String identifier) {
		return options.get(identifier);
	}

	@SuppressWarnings("unchecked")
	public void printUsage() {
		Iterator optionIter = optionsKeys.iterator();
		System.out.println("Usage: Short format, Long format, <Type>, Description -- (Mandatory) [Default value]");
		while(optionIter.hasNext()) {
			ICLIOption option = (ICLIOption) options.get(optionIter.next());
			System.out.println("  "+formatString(shortPrefix, option.getShortLabel(), "")
					+ "\t" + extendToMaxLength(formatString(longPrefix, option.getLongLabel(), ""), maxLongLength)
					+ "\t<" + extendToMaxLength(option.getType()+">", maxTypeLength)
					+ "\t" + option.getDescription()
					+ " -- " + formatMandatory(option.isMandatory())
					+ " " + formatDefault(option.getDefault())
					);
		}
	}
		
	private String formatString(String prefix, String s, String substitution) {
		if(s==null)
			return substitution;
		return prefix+s;
	}
	
	private String formatMandatory(boolean mandatory) {
		if(mandatory)
			return "(mandatory)";
		else
			return "(optional)";
	}
	
	private String formatDefault(String def) {
		if(def == null)
			return "";
		else if(def == "true")
			return "[true]";
		else if(def == "false")
			return "[false]";
		else
			return "["+def+"]";
	}
	
	@SuppressWarnings("unchecked")
	public void addOption(ICLIOption option) {
		if(option.isMandatory())
			mandatory.add(option);
		options.put(longPrefix+option.getLongLabel(), option);
		optionsKeys.add(longPrefix+option.getLongLabel());
		if((option.getShortLabel() != null) && (option.getShortLabel() != ""))
			options.put(shortPrefix+option.getShortLabel(), option);
		if((longPrefix + option.getLongLabel()).length() > maxLongLength)
			maxLongLength = (longPrefix + option.getLongLabel()).length();
		if(option.getType().length() > maxTypeLength)
			maxTypeLength = option.getType().length();
	}
	
	private String extendToMaxLength(String s, int length) {
		int diff = length - s.length();
		for(int i = 0; i < diff; i++)
			s = s+" ";
		return s;
	}

	public void setLongPrefix(String prefix) {
		longPrefix = prefix;
	}

	public void setShortPrefix(String prefix) {
		shortPrefix = prefix;
	}

	public String[] getUnmatched() {
		return (String[]) unmatched.toArray();
	}
	
}
