/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package core.signals;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import core.signals.exceptions.InvalidCallbackException;
import core.signals.exceptions.InvalidSignalException;

public class SignalEmitter {
	
	@SuppressWarnings("unchecked")
	private Class[] signature = {SignalEmitter.class, Object.class};
	private HashMap<String,List<Callback>> callbacks;
	private HashMap<String,String> descriptions;
	
	private class Callback {
		public Object handler;
		public Method cbMethod;
		
		public Callback(Object handler, Method m) {
			this.handler = handler;
			this.cbMethod = m;
		}
		
		public boolean equals(Object o) {
			return o.hashCode() == hashCode();
		}
		
		public int hashCode() {
			return (handler.toString() + cbMethod.getName()).hashCode();
		}
		
		public Object invoke (Object[] args) throws Throwable {
			try {
				return cbMethod.invoke(handler, args);
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			} catch (InvocationTargetException e) {
				throw e.getCause();
			}
			return args;
		}
	}
	
	public SignalEmitter() {
		callbacks = new HashMap<String,List<Callback>>();
		descriptions = new HashMap<String,String>();
	}
	
	/**
	 * Declares a signal (meaning that the class can emit this kind of signal)
	 * @param signalName
	 * @throws DuplicateSignalDeclarationException
	 */
	protected void declareSignal (String signalName, String description) {
		List<Callback> cblist = callbacks.get(signalName);
		if (cblist == null) {
			cblist = new LinkedList<Callback>();
			callbacks.put(signalName, cblist);
		}
		descriptions.put(signalName, description);
	}
	
	/**
	 * Returns the list of declared signals
	 * @return
	 */
	public String[] getSignals () {
		return (String[]) callbacks.keySet().toArray();
	}
	
	/**
	 * Returns a description for the given signal name
	 * @param signalName
	 * @return
	 */
	public String getSignalDescription(String signalName) {
		return descriptions.get(signalName);
	}
	
	/**
	 * Connects a signal to an handler method
	 * @param signalName	Signal name
	 * @param handler		Object that will handle the signal
	 * @param callbackName	Method in the handler that will receive the signal 
	 * 						prototype is: method (SignalEmitter source, Object s);
	 * 						s is the data passed to the callback.
	 * @throws InvalidCallbackException
	 * @throws InvalidSignalException 
	 */
	public void connect (String signalName, Object handler, String callbackName) throws InvalidCallbackException, InvalidSignalException {
		Method m;
		try {
			m = handler.getClass().getMethod(callbackName, signature);
		} catch (SecurityException e) {
			throw new InvalidCallbackException(callbackName);
		} catch (NoSuchMethodException e) {
			throw new InvalidCallbackException(callbackName);
		}
		List<Callback> cblist = callbacks.get(signalName);
		if (cblist == null)
			throw new InvalidSignalException(signalName);
		Callback cb = new Callback(handler, m);
		cblist.add(cb);
	}
	
	/**
	 * Disconnects a signal 
	 * @param signalName
	 * @param handler
	 * @param callbackName
	 * @throws InvalidCallbackException
	 */
	public void disconnect (String signalName, Object handler, String callbackName) throws InvalidCallbackException {
		Method m;
		try {
			m = handler.getClass().getMethod(callbackName, signature);
		} catch (SecurityException e) {
			throw new InvalidCallbackException(callbackName);
		} catch (NoSuchMethodException e) {
			throw new InvalidCallbackException(callbackName);
		}
		List<Callback> cblist = callbacks.get(signalName);
		if (cblist == null)
			return;
		Callback cb = new Callback(handler, m);
		cblist.remove(cb);
	}
	
	/**
	 * Emits a signal stating it comes from source
	 * @param s
	 * @throws InvalidSignalException 
	 * @throws InvocationTargetException 
	 */
	public void delegateEmit (Object source, String signalName, Object s) throws InvalidSignalException, Throwable {
		List<Callback> cblist = callbacks.get(signalName);
		Object[] args = {source, s};
		if (cblist == null)
			throw new InvalidSignalException(signalName);
		for(Callback cb : cblist) {
			cb.invoke(args);
		}	
	}
	
	/**
	 * Emits a signal
	 * @param s
	 * @throws InvalidSignalException 
	 * @throws InvocationTargetException 
	 */
	public void emit (String signalName, Object s) throws InvalidSignalException, Throwable {
		delegateEmit (this, signalName, s);
	}

	
	/**
	 * Emits a signal stating it comes from source and stores the return values of callbacks in a list
	 * @param s
	 * @return
	 * @throws InvalidSignalException 
	 * @throws InvocationTargetException 
	 */
	public List<Object> delegateEmitv (Object source, String signalName, Object s) throws InvalidSignalException, Throwable {
		List<Callback> cblist = callbacks.get(signalName);
		List<Object> retValues = new LinkedList<Object>();
		Object[] args = {source, s};
		if (cblist != null) {
			for(Callback cb : cblist) {
				retValues.add(cb.invoke(args));
			}
		} else
			throw new InvalidSignalException(signalName);
		return retValues;
	}
	
	/**
	 * Emits a signal and stores the return values of callbacks in a list
	 * @param s
	 * @return
	 * @throws InvalidSignalException 
	 * @throws InvocationTargetException 
	 */
	public List<Object> emitv (String signalName, Object s) throws InvalidSignalException, Throwable {
		return delegateEmitv (this, signalName, s);
	}

	
	/**
	 * Emits a signal stating it comes from source and stores the return value of the last callback
	 * @param s
	 * @return
	 * @throws InvalidSignalException 
	 * @throws InvocationTargetException 
	 */
	public Object delegateEmitsinglev (Object source, String signalName, Object s) throws InvalidSignalException, Throwable {
		List<Callback> cblist = callbacks.get(signalName);
		Object retVal = null;
		Object[] args = {source, s};
		if (cblist != null) {
			for(Callback cb : cblist) {
				retVal = cb.invoke(args);
			}
		} else
			throw new InvalidSignalException(signalName);
		return retVal;
	}
	
	/**
	 * Emits a signal and stores the return value of the last callback
	 * @param s
	 * @return
	 * @throws InvalidSignalException 
	 * @throws InvocationTargetException 
	 */
	public Object emitsinglev (String signalName, Object s) throws InvalidSignalException, Throwable {
		return delegateEmitsinglev (this, signalName, s);
	}

	
		
	
	

}
