/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package core.vm;

import com.fuzzbug.solenopsis.sdk.softthreading.IInterruptible;
import com.fuzzbug.solenopsis.sdk.vm.IVM;
import com.fuzzbug.solenopsis.sdk.vm.IVMCode;
import com.fuzzbug.solenopsis.sdk.vm.IVMStack;
import com.fuzzbug.solenopsis.sdk.vm.IVMState;

import core.signals.SignalEmitter;
import core.vm.lib.VMOpUnits;
import core.vm.lib.VMState;

public final class VM extends SignalEmitter implements IVM {
	
	private 	int 		pc = 0;
	private 	int		 	state = IInterruptible.RUNNABLE;
	private		Object		stateMutex = new Object();
	private 	IVMCode 	vmcode;
	private 	IVMStack 	vmstack;
	private		String	 	identifier;
	
	public VM (IVMState state, String id) {
		declareSignal("request-sync", "The VM requests a sync");
		declareSignal("request-stop", "The VM requests a stop");
		declareSignal("extern-op-call", "The VM requests a call to an external operation");
		vmcode = state.getCode();
		pc = state.getPC();
		vmstack = state.getStack();
		this.identifier = id;
	}
	
	public void run() {
		if (!resume()) return; // If we are runnable we can continue, otherwise not
		while (state == IInterruptible.RUNNING) {
			try {
				step();
			} catch (Throwable e) {
				e.printStackTrace();
				stop();
				break;
			}
		}
		runnable(); // Here we become again runnable
		Thread.yield();
	}

	public void step() throws Throwable {
		try {
			IVMCode code = vmcode;
			if (vmstack.getLambda() != null) { // Executing within lambda object
				code = vmstack.getLambda().getCode();
			}
			VMOpUnits.execute(pc, code.getOp(pc), this, vmstack);
		} catch (Exception e) {
			e.printStackTrace();
			stop();
			throw e;
		} finally {
			pc++;
		}
	}

	protected boolean resume() {
		synchronized(stateMutex) {
			if (state == IInterruptible.RUNNABLE) {
				state = IInterruptible.RUNNING;
				return true;
			}
		}
		return false;
	}
	
	protected void runnable() {
		synchronized(stateMutex) {
			if (state == IInterruptible.SUSPENDED) {
				state = IInterruptible.RUNNABLE;
			}
		}
	}

	public void suspend() {
		synchronized(stateMutex) {
			if (state == IInterruptible.RUNNING) {
				state = IInterruptible.SUSPENDED;
			}
		}
	}
	
	public void stop() {
		synchronized(stateMutex) {
			state = IInterruptible.STOPPED;
		}
	}
	
	public void sleep() {
		synchronized(stateMutex) {
			state = IInterruptible.SLEEPING;
		}
	}
	
	public void wakeup() {
		synchronized(stateMutex) {
			if (state == IInterruptible.SLEEPING) {
				state = IInterruptible.RUNNABLE;
			}
		}
	}
	
	public int getPC () {
		return pc;
	}
	
	public void setPC (int pc) {
		this.pc = pc;
	}

	public IVMStack getStack() {
		return vmstack;
	}
	
	public int getInterruptibleStatus() {
		return state;
	}
	
	public IVMState getState() {
		return new VMState(vmstack, vmcode, pc);
	}

	public IVMCode getCode() {
		return vmcode;
	}
	
	public String getIdentifier() {
		return identifier;
	}

}
