/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package core.vm.lib;


import interfaces.vm.util.ISerializer;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.UUID;

import providers.DefaultResourceProvider;
import providers.VMFactory;
import serialization.DeSerializer;
import serialization.Serializer;

import com.fuzzbug.solenopsis.sdk.exe.INodeDaemon;
import com.fuzzbug.solenopsis.sdk.lib.IInstruction;
import com.fuzzbug.solenopsis.sdk.plugins.IPlugin;
import com.fuzzbug.solenopsis.sdk.scheduling.IAlarmable;
import com.fuzzbug.solenopsis.sdk.vm.IVM;

import core.vm.valuetypes.LambdaValue;
import core.vm.valuetypes.ReferenceValue;
import dot.compiler.Compiler;

public class VMBaseOperations implements IPlugin {

	private static VMBaseOperations instance;
	private HashMap<String, Method> methods;
	@SuppressWarnings("unchecked")
	private Class[] signature = { INodeDaemon.class, IVM.class, List.class };
	
	private String[] tagNames = { "and", "equal?", 
			"floatToInt", ">", ">=", 
			"intToFloat", "isDictionary?", "isLambda?", 
			"isList?", "isNil?", "isNumber?", "isRef?", 
			"isString?", "<", "<=", "%",
			"not", "notEqual?", "or", 
			"service?", "get", 
			"hasKey?", "keys", "dict", "put",
			"remove", "allocate", "car", "cdr", 
			"cons", "empty?", "insert", "item", 
			"len", "list", "removeItem", "replace",
			"print", "contains", "=", "string", 
			"!=", "range", "random", "isDefined?", "==", "append", "compare",
			"isSubString?", "clone", "choose", 
			"stringToInt", "stringToFloat", "removeElement", 
			"sleep", "debug", "quit", "exec",
			"getKey", "getKeys", "stringReplace", 
			"stop", "suspend", "randomUUID", "time", 
			"subset", "shuffle", "sort", "floor", "sqrt", "serialize", "deserialize", "evaluate"};

	private String[] methodNames = { "apply_and", "apply_equal",
			"apply_floatToInt", "apply_greater", "apply_greaterEqual",
			"apply_intToFloat", "apply_isDictionary", "apply_isLambda",
			"apply_isList", "apply_isNil", "apply_isNumber", "apply_isRef",
			"apply_isString", "apply_lesser", "apply_lesserEqual", "apply_mod",
			"apply_not", "apply_notEqual", "apply_or",
			"apply_service", "apply_get",
			"apply_hasKey", "apply_keys", "apply_dict", "apply_put",
			"apply_remove", "apply_allocate", "apply_car", "apply_cdr",
			"apply_cons", "apply_empty", "apply_insert", "apply_item",
			"apply_len", "apply_list", "apply_removeItem", "apply_replace",
			"apply_print", "apply_contains", "apply_equal", "apply_string",
			"apply_notEqual", "apply_range", "apply_random", "apply_isdefined",
			"apply_equal", "apply_append", "apply_compare",
			"apply_issubstring", "apply_clone", "apply_choose",
			"apply_stringToInt", "apply_stringToFloat", "apply_removeElement",
			"apply_sleep", "apply_debug", "apply_quit", "apply_exec",
			"apply_getKey", "apply_getKeys", "apply_stringReplace", 
			"apply_stop", "apply_suspend", "apply_randomUUID", "apply_time", 
			"apply_subset",	"apply_shuffle", "apply_sort" ,"apply_floor", "apply_sqrt", 
			"apply_serialize", "apply_deserialize", "apply_evaluate"};

	private Random randomgen;
	
	public static IPlugin instance() {
		if (instance == null) {
			instance = new VMBaseOperations();
		}
		return instance;
	}

	public VMBaseOperations() {
		methods = new HashMap<String, Method>();
		registerMethods(tagNames, methodNames);
		long seed = System.currentTimeMillis();
		randomgen = new Random();
		randomgen.setSeed(seed);
		
		try {
			VMOpUnits.init();
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		}
	}

	private void registerMethods(String[] tags, String[] applyMethods) {
		for (int i = 0; i < tags.length; i++) {
			Method m;
			String methodName = applyMethods[i];
			String tag = tags[i];
			try {
				m = this.getClass().getMethod(methodName, signature);
				methods.put(tag, m);
			} catch (SecurityException e) {
				e.printStackTrace();
			} catch (NoSuchMethodException e) {
				e.printStackTrace();
			}
		}
	}

	public Method getMethod(String tag) {
		return methods.get(tag);
	}

	public boolean isShareable() {
		return true;
	}

	public String getNamespaceName() {
		return "core";
	}

	public String[] getTagsNames() {
		return tagNames;
	}

	public String getAuthor() {
		return "Amos Brocco <amos.brocco@unifr.ch>";
	}

	public String getDescription() {
		return "Solenopsis2: Base Operations Module";
	}

	public String getVersionString() {
		return "2007.11.07";
	}

	public Object apply(Method m, INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Object result = null;
		Object[] fullArgs = new Object[2];
		fullArgs[0] = node;
		fullArgs[1] = args;
		result = m.invoke(this, fullArgs);
		return result;
	}

	public void doInitialize(INodeDaemon node) throws Exception {
	}

	public void doFinalize() throws Exception {
	}

	/* Private utility methods */

	private double getDoubleValue(Object obj) throws Exception {
		if (obj instanceof Number)
			return ((Number) obj).doubleValue();
		else
			return 0;
	}

	private boolean sameType(Object first, Object current) throws Exception {
		if ((first instanceof List<?>) && (current instanceof List<?>))
			return true;
		if ((first instanceof Map<?, ?>) && (current instanceof Map<?, ?>))
			return true;
		if (first.getClass().equals(current.getClass()))
			return true;
		else
			return false;
	}

	/* Method definition */

	public Object apply_and(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		for (int i = 0; i < args.size(); i++) {
			Object current = (Object) args.get(i);
			if (current.equals(false))
				return false;
		}
		return 1;
	}


	public Object apply_equal(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Object first = (Object) args.get(0);
		for (int i = 1; i < args.size(); i++) {
			Object current = (Object) args.get(i);

			// if ((Boolean.FALSE.equals(first))
			// && (current instanceof List)
			// && (((List) current).size() != 0))
			// return false;
			//			
			// if ((Boolean.FALSE.equals(current))
			// && (first instanceof List)
			// && (((List) first).size() != 0))
			// return false;

			if (!(sameType(first, current))) {
				// System.out.println("not same type> " +i + " " +
				// first.getClass()+" " + current.getClass());
				return false;
			}
			if (!(first.hashCode() == current.hashCode())) {
				// if (first instanceof String){
				// String u = (String)first;
				// String v = (String)current;
				// System.out.println("-----"+u+" "+v);
				// for (int j = u.length() - 1; j > 0; j--) {
				// System.out.print((int)u.charAt(j) + " ");
				// }
				// System.out.println();
				// for (int j = v.length() - 1; j > 0; j--) {
				// System.out.print((int)v.charAt(j) + " ");
				// }
				// System.out.println();
				// }
				// System.out.println("not same hashcode> " +i + " *" + (String)
				// first+"* *" + (String) current+"*");
				return false;
			}
		}
		return 1;
	}

	public Object apply_floatToInt(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Object value = args.get(0);
		return new Integer(((Float) value).intValue());
	}

	public Object apply_stringToInt(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		String value = (String) args.get(0);
		return Integer.parseInt(value);
	}

	public Object apply_stringToFloat(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		String value = (String) args.get(0);
		return Float.parseFloat(value);
	}
	
	public Object apply_floor(INodeDaemon node, IVM vm, List<Object> args)
	throws Exception {
		if (args.get(0) instanceof Integer){
			return (Integer) args.get(0);
		} else{
			Double value = (Double) args.get(0);
			return (int) Math.floor(value);
		}
		
		
}

	public Object apply_greater(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		double prec = getDoubleValue(args.get(0));
		for (int i = 1; i < args.size(); i++) {
			double current = getDoubleValue(args.get(i));
			if (prec <= current)
				return false;
			prec = current;
		}
		return 1;
	}

	public Object apply_greaterEqual(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		double prec = getDoubleValue(args.get(0));
		for (int i = 1; i < args.size(); i++) {
			double current = getDoubleValue(args.get(i));
			if (prec < current)
				return false;
			prec = current;
		}
		return 1;
	}

	public Object apply_intToFloat(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Object value = args.get(0);
		return new Float(((Integer) value).intValue());
	}

	public Object apply_isDictionary(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0) instanceof Map<?, ?>)
			return 1;
		return false;
	}

	public Object apply_isLambda(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0) instanceof LambdaValue)
			return 1;
		return false;
	}

	public Object apply_isList(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0) instanceof List<?>)
			return 1;
		return false;
	}

	public Object apply_isNil(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0).equals(false))
			return 1;
		return false;
	}

	public Object apply_isNumber(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if ((args.get(0) instanceof Integer) || (args.get(0) instanceof Float))
			return 1;
		return false;
	}

	public Object apply_isRef(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0) instanceof ReferenceValue)
			return 1;
		return false;
	}

	public Object apply_isString(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0) instanceof String)
			return 1;
		return false;
	}

	public Object apply_lesser(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		double prec = getDoubleValue(args.get(0));
		for (int i = 1; i < args.size(); i++) {
			double current = getDoubleValue(args.get(i));
			if (prec >= current)
				return false;
			prec = current;
		}
		return 1;
	}

	public Object apply_lesserEqual(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		double prec = getDoubleValue(args.get(0));
		for (int i = 1; i < args.size(); i++) {
			double current = getDoubleValue(args.get(i));
			if (prec > current)
				return false;
			prec = current;
		}
		return 1;
	}

	public Object apply_mod(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (getDoubleValue(args.get(0)) == 0) {
			return 0;
		}

		double tempResult = getDoubleValue(args.get(0));
		for (int i = 1; i < args.size(); i++) {
			double value = getDoubleValue(args.get(i));
			if (value == 0)
				return false;
			tempResult %= value;
		}
		
		return tempResult;
	}
	
	public Object apply_sqrt(INodeDaemon node, IVM vm, List<Object> args)
	throws Exception {
		if (getDoubleValue(args.get(0)) == 0) {
			return 0;
		}
		
		return Math.sqrt(getDoubleValue(args.get(0)));
}

	public Object apply_not(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Object first = args.get(0);
		if (first.equals(false))
			return 1;
		else
			return false;
	}

	public Object apply_notEqual(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Object first = (Object) args.get(0);
		for (int i = 1; i < args.size(); i++) {
			Object current = args.get(i);
			if (first.equals(current))
				return false;
			/*
			 * if (sameType(first, current)) return false; if
			 * (first.hashCode()==current.hashCode()) return false;
			 */
		}
		return 1;
	}

	public Object apply_or(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		for (int i = 0; i < args.size(); i++) {
			Object current = (Object) args.get(i);
			if (!(current.equals(false)))
				return 1;
		}
		return false;
	}

	public Object apply_service(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		String name = (String) args.get(0);
		return node.hasMethod(name);
	}

	@SuppressWarnings("unchecked")
	public Object apply_get(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Map dict = (Map) args.get(0);
		String key = (String) args.get(1);
		Object value = dict.get(key);
		if (value == null)
			return false;
		else
			return value;
	}

	@SuppressWarnings("unchecked")
	public Object apply_hasKey(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Map dict = (Map) args.get(0);
		String key = (String) args.get(1);
		if (!dict.containsKey(key))
			return false;
		else
			return 1;
	}

	@SuppressWarnings("unchecked")
	public Object apply_getKey(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Map<String, Object> dict = (Map<String, Object>) args.get(0);
		Object value = args.get(1);
		for (String k : dict.keySet()) {
			Object v = dict.get(k);
			if (v.equals(value)) {
				return k;
			}
		}
		return false;
	}

	@SuppressWarnings("unchecked")
	public Object apply_getKeys(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Map<String, Object> dict = (Map<String, Object>) args.get(0);
		Object value = args.get(1);
		List<String> keys = new ArrayList<String>();
		for (String k : dict.keySet()) {
			Object v = dict.get(k);
			if (v.equals(value)) {
				keys.add(k);
			}
		}
		return keys;
	}

	@SuppressWarnings("unchecked")
	public Object apply_keys(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (Boolean.FALSE.equals(args.get(0)))
			return new LinkedList<String>();
		LinkedList<String> l = new LinkedList<String>();
		Map<String, Object> dict = (Map<String, Object>) args.get(0);
		l.addAll(dict.keySet());
		return l;
	}

	@SuppressWarnings("unchecked")
	public Object apply_new(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.size() < 2)
			return new HashMap();
		List keys = (List) args.get(0);
		List values = (List) args.get(1);

		if (keys.size() != values.size())
			return false;
		HashMap<String, Object> tmpDict = new HashMap<String, Object>();
		for (int i = 0; i < keys.size(); i++) {
			if (!(keys.get(i) instanceof String))
				return false;
			String key = (String) keys.get(i);
			tmpDict.put(key, values.get(i));
		}
		return tmpDict;
	}

	@SuppressWarnings("unchecked")
	public Object apply_put(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Map<String, Object> dict = (Map<String, Object>) args.get(0);
		String key = (String) args.get(1);
		Object value = args.get(2);
		dict.put(key, value);
		return dict;
	}

	@SuppressWarnings("unchecked")
	public Object apply_remove(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Map<String, Object> dict = (Map<String, Object>) args.get(0);
		String key = (String) args.get(1);
		Object result = dict.remove(key);
		if (result == null)
			return false;
		return result;
	}

	public Object apply_allocate(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		int size = (Integer) args.get(0);
		Object item = (Object) args.get(1);
		LinkedList<Object> newList = new LinkedList<Object>();
		for (int i = 0; i < size; i++)
			newList.add(item);
		return newList;
	}

	@SuppressWarnings("unchecked")
	public Object apply_car(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0).equals(false))
			return false;
		List list = (List) args.get(0);
		if (list.size() > 0)
			return list.get(0);
		else
			return false;
	}

	@SuppressWarnings("unchecked")
	public Object apply_cdr(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0).equals(false))
			return false;
		List list = (List) args.get(0);
		if (list.size() > 0) {
			return list.subList(1, list.size());
		} else
			return false;
	}

	@SuppressWarnings("unchecked")
	public Object apply_cons(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		List<Object> list = (List<Object>) args.get(0);
		Object item = (Object) args.get(1);
		list.add(item);
		return list;
	}

	@SuppressWarnings("unchecked")
	public Object apply_empty(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0) instanceof List) {
			List list = (List) args.get(0);
			return list.isEmpty();
		} else if (args.get(0) instanceof Map) {
			Map m = (Map) args.get(0);
			return m.isEmpty();
		} else if (args.get(0) == Boolean.FALSE) {
			return 1;
		}
		return false;
	}

	@SuppressWarnings("unchecked")
	public Object apply_insert(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		List list;
		if (Boolean.FALSE.equals(args.get(0))) {
			list = new ArrayList();
		}
		list = (List) args.get(0);
		Integer position = (Integer) args.get(1);
		Object item = args.get(2);
		if ((position < 0) || (position > list.size()))
			return false;
		list.add(position, item);
		return list;
	}

	@SuppressWarnings("unchecked")
	public Object apply_item(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (Boolean.FALSE.equals(args.get(0)))
			return false;
		List list = (List) args.get(0);
		int position = (Integer) args.get(1);
		if ((list.size() > 0) && (position >= 0)
				&& (position <= list.size() - 1))
			return list.get(position);
		else
			return false;
	}

	@SuppressWarnings("unchecked")
	public Object apply_length(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (Boolean.FALSE.equals(args.get(0)))
			return 0;
		List list = (List) args.get(0);
		return list.size();
	}

	public Object apply_list(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		LinkedList<Object> tmpList = new LinkedList<Object>();
		for (int i = 0; i < args.size(); i++)
			tmpList.add(args.get(i));
		return tmpList;
	}

	@SuppressWarnings("unchecked")
	public Object apply_dict(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.size() < 2)
			return new HashMap();
		List keys = (List) args.get(0);
		List values = (List) args.get(1);

		if (keys.size() != values.size())
			return false;
		Map<String, Object> tmpDict = new HashMap<String, Object>();
		for (int i = 0; i < keys.size(); i++) {
			if (!(keys.get(i) instanceof String))
				return false;
			String key = (String) keys.get(i);
			tmpDict.put(key, values.get(i));
		}

		return tmpDict;
	}

	@SuppressWarnings("unchecked")
	public Object apply_len(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.get(0).equals(false))
			return 0;
		List list = (List) args.get(0);
		return list.size();
	}

	@SuppressWarnings("unchecked")
	public Object apply_removeItem(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		List list = (List) args.get(0);
		int position = (Integer) args.get(1);
		if ((list.size() > 0) && (position >= 0)
				&& (position <= list.size() - 1))
			return list.remove(position);
		else
			return false;
	}

	@SuppressWarnings("unchecked")
	public Object apply_removeElement(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		List list = (List) args.get(0);
		Object element = (Object) args.get(1);
		if (list.contains(element)) {
			if (list.remove(element))
				return list;
		}
		return false;

	}

	@SuppressWarnings("unchecked")
	public Object apply_replace(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		List list = (List) args.get(0);
		int position = (Integer) args.get(1);
		Object item = args.get(2);
		if ((list.size() > 0) && (position >= 0)
				&& (position <= list.size() - 1)) {
			Object old = list.remove(position);
			list.add(position, item);
			return old;
		} else
			return false;
	}

	public Object apply_print(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		System.out.print("(****) ");
		for (int i = 0; i < args.size(); i++)
			System.out.print(args.get(i).toString().replace("\\n", "\n").replace("\\t", "\t"));
		System.out.println();
		return false;
	}

	public Object apply_debug(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		String source;
		try {
			source = (String) vm.getStack().getGValue("$SELF");
		} catch (Exception e) {
			source = "n/a";
		}
		System.out.format("(@@@@) Debug (%s) : ", source);

		for (int i = 0; i < args.size(); i++)
			System.out.print(args.get(i).toString());
		System.out.println();
		return false;
	}

	@SuppressWarnings("unchecked")
	public Object apply_contains(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (Boolean.FALSE.equals(args.get(0)))
			return false;
		List list = (List) args.get(0);
		Object item = args.get(1);
		return list.contains(item);
	}

	public Object apply_string(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		String result = "";
		for (int i = 0; i < args.size(); i++)
			result += args.get(i).toString();
		return result;
	}

	public Object apply_range(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Integer from = (Integer) args.get(0);
		Integer to = (Integer) args.get(1);
		List<Integer> retVal = new ArrayList<Integer>();
		if (from <= to) {
			for (int i = from; i <= to; i++) {
				retVal.add(i);
			}
		} else {
			for (int i = from; i >= to; i--) {
				retVal.add(i);
			}
		}

		return retVal;
	}

	public Object apply_random(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		return randomgen.nextFloat();
	}

	@SuppressWarnings("unchecked")
	public Object apply_choose(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		List<Object> list = (List<Object>) args.get(0);
		if (list.size() > 0) {
			int rnd = (int) (randomgen.nextDouble() * list.size());
			return list.get(rnd);
		}
		return false;
	}
	
	public Object apply_evaluate(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		String opath = (String) args.get(0);
		InputStream is = DefaultResourceProvider.instance().getResourceInputStream(opath);
		List<IInstruction> ilist = Compiler.compile(DefaultResourceProvider.instance(), opath, is);
		VMCode code = new VMCode(opath, ilist);
		VMStack stack = new VMStack();
		VMState st = new VMState(stack, code, 0);
		IVM nvm = VMFactory.getVM(st, node);
		nvm.run();
		Object tos = stack.getTOS();
		node.unschedule(nvm);
		return tos;
	}
	
	@SuppressWarnings("unchecked")
	public Object apply_shuffle(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		List<Object> list = (List<Object>) args.get(0);
		List<Object> copy = new ArrayList<Object>();
		copy.addAll(list);
		Collections.shuffle(copy);
		return copy;
	}
	
	@SuppressWarnings("unchecked")
	public Object apply_subset (INodeDaemon node, IVM vm,  List<Object> args) throws Exception {
		List<Object> list = (List<Object>) args.get(0);
		Number count = (Number) args.get(1);
		if (count.intValue() > list.size()) return list;
		ArrayList<Object> tmp = new ArrayList<Object>();
		tmp.addAll(list);
		Collections.shuffle(tmp);
		return tmp.subList(0, count.intValue()-1);
	}
	
	@SuppressWarnings("unchecked")
	public Object apply_sort(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		try{
			List<Comparable> list = (List<Comparable>) args.get(0);
			List<Comparable> copy = new ArrayList<Comparable>();
			copy.addAll(list);
			Collections.sort(copy);
			return copy;
		} catch (Exception e){
			return args.get(0);
		}
	}

	public Object apply_isdefined(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		return args.get(0) != null ? 1 : false;
	}

	@SuppressWarnings("unchecked")
	public Object apply_append(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		List list = (List) args.get(0);
		synchronized (list) {
			for (int i = 1; i < args.size(); i++) {
				Object item = args.get(i);
				list.add(item);
			}
		}
		return list;
	}

	@SuppressWarnings("unchecked")
	private boolean compare(Object first, Object second) {
		try {
			if (!(sameType(first, second))) {
				return false;
			}
		} catch (Exception e) {
			return false;
		}
		if (first instanceof List) {
			List flist = (List) first;
			List slist = (List) second;
			if (flist.size() != slist.size())
				return false;
			for (Object i : flist) {
				if (!slist.contains(i)) {
					return false;
				}
			}
			return true;
		} else if (first instanceof Map) {
			Map fmap = (Map) first;
			Map smap = (Map) second;
			if (fmap.size() != smap.size())
				return false;
			for (Object i : fmap.keySet()) {
				if (!smap.containsKey(i))
					return false;
				if (!(compare(fmap.get(i), smap.get(i))))
					return false;
			}
			return true;
		} else {
			if (first.equals(second))
				return true;
			else
				return false;
		}
	}

	public Object apply_compare(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		if (args.size() == 1) {
			return 1;
		}
		if (args.size() != 2) {
			return false;
		}
		Object first = (Object) args.get(0);
		Object second = (Object) args.get(1);
		if (compare(first, second))
			return 1;
		else
			return false;

	}

	@SuppressWarnings("unchecked")
	public Object apply_exec(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {

		Map<String, String> variables = System.getenv();
		List<String> envp = new ArrayList<String>();
		for (String t : variables.keySet()) {
			envp.add(t + "=" + variables.get(t));
		}
		if (args.size() > 1) {
			List<String> env = (List<String>) args.get(1);
			for (String t : env) {
				envp.add(t);
			}
		}

		final String[] envar = envp.toArray(new String[envp.size()]);
		// System.out.println("(INFO) Environment variables are "+envp);
		try {
			final String command = (String) args.get(0);
			System.out
					.print("(INFO) Spawning new thread for external command '"
							+ command + "'");
			new Thread() {
				public void run() {
					try {
						Process p = Runtime.getRuntime().exec(command, envar);
						BufferedReader is = new BufferedReader(
								new InputStreamReader(p.getInputStream()));
						String line;
						while ((line = is.readLine()) != null)
							System.out.println(line);

					} catch (IOException e) {
						e.printStackTrace();
					}
				}
			}.start();
			System.out.println("done");
			return 1;
		} catch (Exception e) {
			System.out.println("failed");
			return false;
		}
	}

	public Object apply_issubstring(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		String first = (String) args.get(0);
		String second = (String) args.get(1);
		if (first.contains(second))
			return 1;
		else
			return false;
	}

	public Object apply_stringReplace(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		String first = (String) args.get(0);
		String target = (String) args.get(1);
		String replacement = (String) args.get(2);
		return first.replace(target, replacement);
	}

	public Object apply_sleep(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		Integer time = (Integer) args.get(0);
		Long t = new Long(time);
		
		final IVM sleepingBeauty = vm;
		IAlarmable alarm = new IAlarmable() {
			@Override
			public void callback(UUID taskId, Object data, long timestamp) {
				((IVM) data).wakeup();
			}
			
		};
		vm.sleep();
		node.getNodeAlarmService().scheduleAfter(alarm, sleepingBeauty, t);
		Thread.yield();
		return false;
	}
	
	public Object apply_time (INodeDaemon node, IVM vm,  List<Object> args) throws Exception {
		return System.currentTimeMillis();
	}
	
	public Object apply_stop (INodeDaemon node, IVM vm,  List<Object> args) throws Exception {
		vm.stop();
		return false;
	}
	
	public Object apply_suspend (INodeDaemon node, IVM vm,  List<Object> args) throws Exception {
		vm.suspend();
		return false;
	}
	
	public Object apply_randomUUID (INodeDaemon node, IVM vm,  List<Object> args) throws Exception {
		return UUID.randomUUID().toString();
	}
	

	public void apply_quit(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		System.exit(0);
	}

	public Object apply_clone(INodeDaemon node, IVM vm, List<Object> args)
			throws Exception {
		return deepClone(args.get(0));
	}
	
	public Object apply_serialize(INodeDaemon node, IVM vm, List<Object> args)
	throws Exception {
			Object value = args.get(0);
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			ISerializer ser = new Serializer(out);
			try {
				ser.print(value);
			} catch (Exception e) {
				e.printStackTrace();
			}
			out.flush();
			String state = out.toString();
			out.close();
			return state;
	}
	
	public Object apply_deserialize(INodeDaemon node, IVM vm, List<Object> args)
	throws Exception {
			String value = (String) args.get(0);
			
			ByteArrayInputStream bais = new ByteArrayInputStream(value.getBytes());
			Object data = false;
			DeSerializer ds = null;
			ds = new DeSerializer(bais);
			bais.close();
			try {
				data = ds.get();
			} catch (Exception e) {
				return false;
			}
			return data;
	}

	@SuppressWarnings("unchecked")
	private Object deepClone(Object o) throws Exception {
		if (o instanceof List) {
			List<Object> l = new LinkedList<Object>();
			List origin = (List) o;
			for (Object item : origin) {
				l.add(deepClone(item));
			}
			return l;
		} else if (o instanceof Map) {
			Map<String, Object> m = new HashMap<String, Object>();
			Map<String, Object> origin = (Map<String, Object>) o;
			for (String key : origin.keySet()) {
				m.put(new String(key), deepClone(origin.get(key)));
			}
			return m;
		} else if (o instanceof String) {
			return new String((String) o);
		} else if (o instanceof Integer) {
			return new Integer((Integer) o);
		} else if (o instanceof Float) {
			return new Float((Float) o);
		} else if (o instanceof Boolean) {
			return new Boolean((Boolean) o);
		} else if (o instanceof LambdaValue) {
			return ((LambdaValue) o).clone();
		} else
			return o;
	}

}
