/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package core.vm.lib;

import interfaces.vm.valuetypes.ILambdaValue;
import interfaces.vm.valuetypes.IReferenceValue;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import util.Dolly;

import com.fuzzbug.solenopsis.sdk.lib.IConstants;
import com.fuzzbug.solenopsis.sdk.lib.IInstruction;
import com.fuzzbug.solenopsis.sdk.vm.IVMStack;

import core.signals.exceptions.InvalidSignalException;
import core.signals.signals.ExternFunctionCallSignalData;
import core.signals.signals.StopRequestSignalData;
import core.signals.signals.SyncRequestSignalData;
import core.vm.VM;
import core.vm.exceptions.InvalidOpcodeException;
import core.vm.exceptions.InvalidTypeException;
import core.vm.valuetypes.LambdaValue;
import core.vm.valuetypes.ReferenceValue;
import dot.codegenerator.Instruction;

@SuppressWarnings("unchecked")
public class VMOpUnits {
	
	public static Method[] mops;
	
	public static void init() throws SecurityException, NoSuchMethodException {
		mops = new Method[IConstants.OPCODES.values().length];	
		mops[IConstants.OPCODES.LOAD.ordinal()] = VMOpUnits.class.getMethod("load", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.GLOAD.ordinal()] = VMOpUnits.class.getMethod("gload", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.STORE.ordinal()] = VMOpUnits.class.getMethod("store", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.GSTORE.ordinal()] = VMOpUnits.class.getMethod("gstore", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.SET.ordinal()] = VMOpUnits.class.getMethod("set", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.GSET.ordinal()] = VMOpUnits.class.getMethod("gset", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.POP.ordinal()] = VMOpUnits.class.getMethod("pop", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.DUP.ordinal()] = VMOpUnits.class.getMethod("dup", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.CONST.ordinal()] = VMOpUnits.class.getMethod("cconst", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.JUMP.ordinal()] = VMOpUnits.class.getMethod("jump", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.FJUMP.ordinal()] = VMOpUnits.class.getMethod("fjump", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.BIND.ordinal()] = VMOpUnits.class.getMethod("bind", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.RETURN.ordinal()] = VMOpUnits.class.getMethod("rreturn", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.NOP.ordinal()] = VMOpUnits.class.getMethod("nop", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.LABEL.ordinal()] = VMOpUnits.class.getMethod("label", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.FRAME.ordinal()] = VMOpUnits.class.getMethod("frame", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.END.ordinal()] = VMOpUnits.class.getMethod("end", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.DROP.ordinal()] = VMOpUnits.class.getMethod("drop", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.SYNC.ordinal()] = VMOpUnits.class.getMethod("sync", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.LAMBDA.ordinal()] = VMOpUnits.class.getMethod("lambda", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.RLOAD.ordinal()] = VMOpUnits.class.getMethod("rload", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.URLOAD.ordinal()] = VMOpUnits.class.getMethod("urload", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.ICALL.ordinal()] = VMOpUnits.class.getMethod("icall", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.ECALL.ordinal()] = VMOpUnits.class.getMethod("ecall", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.LOCAL.ordinal()] = VMOpUnits.class.getMethod("local", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.LCALL.ordinal()] = VMOpUnits.class.getMethod("lcall", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.ITEM.ordinal()] = VMOpUnits.class.getMethod("item", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.INC.ordinal()] = VMOpUnits.class.getMethod("inc", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.LEN.ordinal()] = VMOpUnits.class.getMethod("len", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.LT.ordinal()] = VMOpUnits.class.getMethod("lt", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.EQ.ordinal()] = VMOpUnits.class.getMethod("eq", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.GET.ordinal()] = VMOpUnits.class.getMethod("get", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.STRING.ordinal()] = VMOpUnits.class.getMethod("string", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.ADD.ordinal()] = VMOpUnits.class.getMethod("add", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.LIST.ordinal()] = VMOpUnits.class.getMethod("list", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.DICT.ordinal()] = VMOpUnits.class.getMethod("dict", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.SUB.ordinal()] = VMOpUnits.class.getMethod("sub", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.MUL.ordinal()] = VMOpUnits.class.getMethod("mul", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.DIV.ordinal()] = VMOpUnits.class.getMethod("div", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.PUT.ordinal()] = VMOpUnits.class.getMethod("put", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.APPEND.ordinal()] = VMOpUnits.class.getMethod("append", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.PC.ordinal()] = VMOpUnits.class.getMethod("pc", Integer.class, IInstruction.class, VM.class, IVMStack.class);
		mops[IConstants.OPCODES.UNWIND.ordinal()] = VMOpUnits.class.getMethod("unwind", Integer.class, IInstruction.class, VM.class, IVMStack.class);
	}
	
	
	public static void execute(Integer pc, IInstruction fullop, VM vm, IVMStack stack)
			throws IndexOutOfBoundsException, InvalidSignalException,
			InvalidOpcodeException, InvalidTypeException, Throwable {
		try {
			//System.out.println(pc+"->"+fullop);
			mops[fullop.getOpCode().ordinal()].invoke(null, pc, fullop, vm, stack);
		} catch (Throwable t) {
			t.printStackTrace();
			throw t.getCause();
		}	
	}
	

	static List toList(Object obj) throws Throwable {
		if (obj instanceof List)
			return (List) obj;
		else if (Boolean.FALSE.equals(obj))
			return new ArrayList();
		else
			throw new Exception("Incompatible type for dictionary (expected List or Nil, but got "+obj.getClass()+")");
	}
	

	private static Map toDict(Object obj) throws Throwable {
		if (obj instanceof Map)
			return (Map) obj;
		else if (Boolean.FALSE.equals(obj))
			return new HashMap<String,Object>();
		else
			throw new Exception("Incompatible type for dictionary (expected Nil, but got "+obj.getClass()+")");
	}
	
	static Number asNumber(Object x) {
		if (x instanceof Number) {
			return (Number) x;
		} return 0;
	}
	
	public static void load(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		stack.pushValue(stack.getValue((String) op.get(0)));
	}

	public static void gload(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		stack.pushValue(stack.getGValue((String) op.get(0)));
	}
	public static void store(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		stack.storeValue((String) op.get(0), stack.popValue());
	}
	public static void gstore(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		stack.storeValue((String) op.get(0), stack.popValue());
	}
	public static void set(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		String vname = (String) op.get(0);
		if (stack.getValue(vname) instanceof IReferenceValue)
			stack.setReferenceValue(vname, stack.popValue());
		else
			stack.setValue(vname, stack.popValue());
	}
	public static void gset(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		stack.setGValue((String) op.get(0), stack.popValue());
	}
	public static void pop(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		stack.popValue();
	}
	public static void dup(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		Object tos = stack.getTOS();
		stack.pushValue(tos);
	}
	public static void cconst(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		try {
			stack.pushValue((Object) Dolly.deepClone(op.get(0)));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	public static void jump(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		vm.setPC(pc + ((Integer) op.get(0)).intValue() - 1);
	}
	public static void fjump(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		if (stack.popValue().equals(false))
			vm.setPC(pc + ((Integer) op.get(0)).intValue() - 1);
	}
	public static void bind(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		List bindNames = (List) op.get(0);
		for (int i = bindNames.size() - 1; i >= 0; i--) {
			stack.storeValue((String) bindNames.get(i), stack.popValue());
		}
		if (op.size() > 1) {
			List bindNilNames = (List) op.get(1);
			for (int i = bindNilNames.size() - 1; i >= 0; i--) {
				stack.setValue((String) bindNilNames.get(i), null);
			}
		}
	}
	public static void rreturn(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		// Save the return value
		Object returnValue = false;
		if (!stack.isEmpty()) {
			returnValue = stack.popValue();
		}
		// Remove all local environments
		while (stack.isLocalStackFrame()) {
			stack.popStackFrame();
		}
		// Retrieve return pc
		int returnPC = stack.getReturnPC();
		if (stack.getLambda() != null && stack.getLambda().getLambdaStart().intValue() != -1)
			stack.getLambda().updateBindings(stack);
		// Pop the function frame
		stack.popStackFrame();
		stack.pushValue(returnValue);
		vm.setPC(returnPC);
	}
	public static void nop(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		return;
	}
	public static void label(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		return;
	}
	public static void frame(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		return;
	}
	public static void end(Integer pc, IInstruction fullop, VM vm, IVMStack stack) throws InvalidSignalException, Throwable {
		vm.setPC(-1);
		vm.emit("request-stop", new StopRequestSignalData());
	}
	
	public static void drop(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		// Save the return value
		Object retVal = false;
		if (!stack.isEmpty()) {
			retVal = stack.popValue();
		}
		stack.popStackFrame();
		stack.pushValue(retVal);
	}
	public static void sync(Integer pc, IInstruction fullop, VM vm, IVMStack stack) throws InvalidSignalException, Throwable {
		vm.emit("request-sync", new SyncRequestSignalData());
	}
	public static void lambda(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		/* New Lambda definition */
		int lambdaAbsStart = pc + ((Integer) op.get(0)).intValue();
		Integer lambdaStart = new Integer(lambdaAbsStart);
		List varNames = (List) op.get(1);
		/* Construct the variable values list */
		Map<String,Object> bindings = new HashMap<String,Object>();
		for (int k = 0; k < varNames.size(); k++) {
			String vname = (String) varNames.get(k);
			Object value = stack.getValue(vname);
			bindings.put(vname, value);
		}
		ILambdaValue lambda = new LambdaValue(vm.getCode(), lambdaStart, bindings);
		stack.pushValue(lambda);
	}
	public static void rload(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		String name = (String) op.get(0);
		stack.pushValue(new ReferenceValue(name, stack.getTopFrameIndex()));
	}
	public static void urload(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		String rname = (String) op.get(0);
		stack.pushValue(stack.getReferenceValue((IReferenceValue) stack
				.getValue(rname)));
	}
	public static void icall(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		List iFargs = stack.popNValues(((Integer) op.get(1)).intValue());
		stack.pushStackFrame(pc);
		stack.pushNValues(iFargs);
		vm.setPC(pc + ((Integer) op.get(0)).intValue() - 1);
	}
	public static void ecall(Integer pc, IInstruction fullop, VM vm, IVMStack stack) throws InvalidSignalException, Throwable {
		List<Object> op = fullop.getParameters();
		String eFunctionName = (String) op.get(0);
		List eFargs = stack.popNValues(((Integer) op.get(1)).intValue());
		stack.pushValue(vm.emitsinglev("extern-op-call",
				new ExternFunctionCallSignalData(eFunctionName, eFargs)));
	}
	public static void local(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		stack.pushLocalStackFrame();
	}
	public static void lcall(Integer pc, IInstruction fullop, VM vm, IVMStack stack) throws InvalidTypeException {
		List<Object> op = fullop.getParameters();
		Object lfsymbol = stack.popValue();
		List lfFargs = stack.popNValues(((Integer) op.get(0)).intValue());
		/* Determine if there is a lambda object associated */
		if ((lfsymbol != null) && (lfsymbol instanceof ILambdaValue)) {
			/* Lambda call */
			ILambdaValue lambda = (ILambdaValue) lfsymbol;
			stack.pushStackFrame(pc, lambda);
			stack.pushNValues(lfFargs);
			/* Define saved lambda arguments */
			Map<String, Object> bindings = lambda.getBindings();
			Iterator<String> varNames = bindings.keySet().iterator();
			while (varNames.hasNext()) {
				String name = varNames.next();
				stack.storeValue(name, bindings.get(name));
			}
			/* Update pc to jump to correct address */
			vm.setPC(lambda.getLambdaStart().intValue() - 1);
		} else {
			throw new InvalidTypeException(
					"No lambda object bound to variable "+lfsymbol.toString()+" @ line:"+((Instruction)fullop).getNode().getLine()+" position:"+((Instruction)fullop).getNode().getPosition());
		}
	}
	public static void item(Integer pc, IInstruction fullop, VM vm, IVMStack stack) throws Throwable {
		int position = (Integer) stack.popValue();
		List list = toList(stack.popValue());
		if ((list.size() > 0) && (position >= 0)
				&& (position <= list.size() - 1))
			stack.pushValue(list.get(position));
		else
			stack.pushValue(false);
	}
	public static void inc(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		int value = (Integer) stack.popValue();
		stack.pushValue(value+1);
	}
	public static void len(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		if (Boolean.FALSE.equals(stack.getTOS())) {
			stack.popValue();
			stack.pushValue(0);
		} else {
			List list = (List) stack.popValue();
			stack.pushValue(list.size());
		}
	}
	public static void lt(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		Number second = asNumber(stack.popValue());
		Number first = asNumber(stack.popValue());
		stack.pushValue(first.doubleValue() < second.doubleValue() ? 1 : false);
	}
	public static void eq(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		Object v1 = stack.popValue();
		Object v2 = stack.popValue();
		stack.pushValue(v1.equals(v2) ? 1 : false);
	}
	public static void get(Integer pc, IInstruction fullop, VM vm, IVMStack stack) throws Throwable {
		Object key = stack.popValue();
		Map dict = toDict(stack.popValue());
		Object val = dict.get(key);
		stack.pushValue(val == null ? false : val);
	}
	public static void string(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		List strings = stack.popNValues(((Integer) op.get(0)).intValue());
		StringBuilder b = new StringBuilder();
		for (Object pk : strings)
			b.append(pk.toString());
		stack.pushValue(b.toString());
	}
	public static void add(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		Number aavo = asNumber(stack.popValue());
		Number abvo = asNumber(stack.popValue());
		stack.pushValue(VMMath.sum(aavo, abvo));
	}
	public static void list(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		List<Object> op = fullop.getParameters();
		int count = ((Integer) op.get(0)).intValue();
		stack.pushValue(stack.popNValues(count));
	}
	public static void dict(Integer pc, IInstruction fullop, VM vm, IVMStack stack) throws Throwable {
		List values = toList(stack.popValue());
		List keys = toList(stack.popValue());
		if (keys.size() != values.size()) {
			throw new InvalidTypeException("List size are not equal");
		}
		Map<String,Object> tmpDict = new HashMap<String,Object>();
		for (int i = 0; i < keys.size(); i++) {
			if(!(keys.get(i) instanceof String)) {
				throw new InvalidTypeException(
				"Key must be a string");
			}
			tmpDict.put((String) keys.get(i), values.get(i));
		}
		stack.pushValue(tmpDict);
	}
	public static void sub(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		Number savo = asNumber(stack.popValue());
		Number sbvo = asNumber(stack.popValue());
		stack.pushValue(VMMath.sub(savo, sbvo));
	}
	public static void mul(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		Number mavo = asNumber(stack.popValue());
		Number mbvo = asNumber(stack.popValue());
		stack.pushValue(VMMath.mul(mavo, mbvo));
	}
	public static void div(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		Number davo = asNumber(stack.popValue());
		Number dbvo = asNumber(stack.popValue());
		stack.pushValue(VMMath.div(davo, dbvo));
	}
	public static void put(Integer pc, IInstruction fullop, VM vm, IVMStack stack) throws Throwable {
		Object pvalue = stack.popValue();
		Object pkey = stack.popValue();
		Map pdict = toDict(stack.popValue());
		synchronized(pdict) {
			pdict.put(pkey, pvalue);
			stack.pushValue(pdict);
		}
	}
	public static void append(Integer pc, IInstruction fullop, VM vm, IVMStack stack) throws Throwable {
		Object el = stack.popValue();
		List elist = toList(stack.popValue());
		synchronized(elist) {
			elist.add(el);
			stack.pushValue(elist);
		}
	}
	public static void pc(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {

	}
	public static void unwind(Integer pc, IInstruction fullop, VM vm, IVMStack stack) {
		if (!Boolean.FALSE.equals(stack.getTOS())) {
			List tlist = (List) stack.popValue();
			for (Object x : tlist) {
				stack.pushValue(x);
			}
		}
	}



}
