/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package core.vm.lib;

import interfaces.vm.valuetypes.ILambdaValue;
import interfaces.vm.valuetypes.IReferenceValue;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.fuzzbug.solenopsis.sdk.vm.IVMStack;
import com.fuzzbug.solenopsis.sdk.vm.IVMStackFrame;


public class VMStack implements IVMStack {

	private static final long serialVersionUID = 5119570498932934269L;

	private IVMStackFrame currentStackFrame;

	private IVMStackFrame mainStackFrame;

	private List<IVMStackFrame> stackFramesStack;

	public VMStack() {
		stackFramesStack = new LinkedList<IVMStackFrame>();
		mainStackFrame = pushStackFrame(0);
	}
	
	public boolean isEmpty() {
		return currentStackFrame.isEmpty();
	}

	@SuppressWarnings("unchecked")
	public VMStack(List stackRepr) {
		stackFramesStack = new LinkedList<IVMStackFrame>();
		for (int i = 0; i < stackRepr.size(); i++)
			stackFramesStack
					.add(new VMStackFrame((List) stackRepr.get(i)));
		currentStackFrame = (VMStackFrame) stackFramesStack
				.get(stackFramesStack.size() - 1);
		mainStackFrame = (VMStackFrame) stackFramesStack.get(0);
	}

	public Object getGValue(String var) {
		return mainStackFrame.getValue(var);
	}

	public int getReturnPC() {
		return currentStackFrame.getReturnPC();
	}
	
	public ILambdaValue getLambda() {
		return currentStackFrame == null ? null : currentStackFrame.getLambda();
	}

	private IVMStackFrame getStackFrameDefining(String var) {
		Object value = null;
		int i = stackFramesStack.size() - 1;
		VMStackFrame searchedStackFrame = null;
		while ((value == null) && (i >= 0)) {
			searchedStackFrame = (VMStackFrame) stackFramesStack.get(i);
			value = searchedStackFrame.getValue(var);
			i--;
		}
		return searchedStackFrame;
	}

	public Object getTOS() {
		return currentStackFrame.getTOS();
	}

	public Object getValue(String var) {
		IVMStackFrame sf = getStackFrameDefining(var);
		if (sf == null)
			return false;
		else
			return sf.getValue(var);
	}
	
	private IVMStackFrame getStackFrameDefining(String var, int i) {
		Object value = null;
		VMStackFrame searchedStackFrame = null;
		while ((value == null) && (i >= 0)) {
			searchedStackFrame = (VMStackFrame) stackFramesStack.get(i);
			value = searchedStackFrame.getValue(var);
			i--;
		}
		return searchedStackFrame;
	}
	
	public Object getReferenceValue(IReferenceValue ref) {
		String var = ref.getReferencedName();
		Integer index = ref.getFrameIndex();
		IVMStackFrame sf = getStackFrameDefining(var, index.intValue());
		if (sf == null)
			return false;
		else
			return sf.getValue(var);
	}

	@SuppressWarnings("unchecked")
	public List popNValues(int i) {
		if (i < 0) i = currentStackFrame.getSize();
		List tempList = new ArrayList(i);
		for (int k = 0; k < i; k++) {
			tempList.add(0, currentStackFrame.popValue());
		}
		return tempList;
	}

	public IVMStackFrame popStackFrame() {
		currentStackFrame = (IVMStackFrame) stackFramesStack
				.get(stackFramesStack.size() - 2);
		return (IVMStackFrame) stackFramesStack.remove(stackFramesStack.size()-1);
	}

	public Object popValue() {
		return currentStackFrame.popValue();
	}

	public void pushNValues(List<Object> args) {
		for (int k = 0; k < args.size(); k++) {
			currentStackFrame.pushValue(args.get(k));
		}
	}

	public IVMStackFrame pushStackFrame(int pc) {
		return pushStackFrame(pc, getLambda());
	}

	public IVMStackFrame pushStackFrame(int pc, ILambdaValue lambda) {
		currentStackFrame = new VMStackFrame(pc, lambda, false);
		stackFramesStack.add(currentStackFrame);
		return currentStackFrame;
	}

	public void pushValue(Object argument) {
		currentStackFrame.pushValue(argument);
	}

	public void setGValue(String argument, Object value) {
		mainStackFrame.setValue(argument, value);
	}

	public void setValue(String var, Object value) {
		IVMStackFrame sf = getStackFrameDefining(var);
		if (sf == null)
			currentStackFrame.setValue(var, value);
		else
			sf.setValue(var, value);
	}

	public void storeGValue(String argument, Object value) {
		mainStackFrame.setValue(argument, value);
	}

	public void storeValue(String argument, Object value) {
		currentStackFrame.setValue(argument, value);
	}

	public Object toBaseValue() {
		return stackFramesStack;
	}

	public Integer getTopFrameIndex() {
		return new Integer(stackFramesStack.size() - 1);
	}

	public void setReferenceValue(String name, Object value) {
		IReferenceValue ref = (IReferenceValue) getValue(name);
		String var = ref.getReferencedName();
		Integer index = ref.getFrameIndex();
		IVMStackFrame sf = getStackFrameDefining(var, index.intValue());
		if (sf != null)
			sf.setValue(var, value);
	}

	public boolean isLocalStackFrame() {
		return currentStackFrame.isLocal();
	}

	public IVMStackFrame pushLocalStackFrame() {
		currentStackFrame = new VMStackFrame(currentStackFrame.getReturnPC(), getLambda(), true);
		stackFramesStack.add(currentStackFrame);
		return currentStackFrame;
	}

	public Set<String> getGVariables() {
		return mainStackFrame.getVariables();
	}
	
	public String toString() {
		StringBuilder str = new StringBuilder();
		for (IVMStackFrame f : stackFramesStack) {
			str.append(f.toString());
			str.append("\n");
		}
		return str.toString();
	}

	@Override
	public void updateAllLambdaBindings() {
		for (IVMStackFrame f : stackFramesStack) {
			if (f.getLambda() != null) {
				f.getLambda().updateBindings(this);
			}
		}
	}

}
