/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package dot.codegenerator;


import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import com.fuzzbug.solenopsis.sdk.lib.IConstants;
import com.fuzzbug.solenopsis.sdk.lib.IInstruction;

import logging.Logger;
import dot.codegenerator.macros.Macro;
import dot.codegenerator.modules.Add_Generator;
import dot.codegenerator.modules.Append_Generator;
import dot.codegenerator.modules.Begin_Generator;
import dot.codegenerator.modules.Break_Generator;
import dot.codegenerator.modules.Call_Generator;
import dot.codegenerator.modules.Concat_Generator;
import dot.codegenerator.modules.Cond_Generator;
import dot.codegenerator.modules.Context_Generator;
import dot.codegenerator.modules.Continue_Generator;
import dot.codegenerator.modules.Defmacro_Generator;
import dot.codegenerator.modules.Defun_Generator;
import dot.codegenerator.modules.Defvar_Generator;
import dot.codegenerator.modules.Dict_Generator;
import dot.codegenerator.modules.Div_Generator;
import dot.codegenerator.modules.EndFun_Generator;
import dot.codegenerator.modules.End_Generator;
import dot.codegenerator.modules.Equal_Generator;
import dot.codegenerator.modules.For_Generator;
import dot.codegenerator.modules.Foreach_Generator;
import dot.codegenerator.modules.Get_Generator;
import dot.codegenerator.modules.Goto_Generator;
import dot.codegenerator.modules.IGenerator;
import dot.codegenerator.modules.If_Generator;
import dot.codegenerator.modules.Label_Generator;
import dot.codegenerator.modules.Lambda_Generator;
import dot.codegenerator.modules.Let_Generator;
import dot.codegenerator.modules.List_Generator;
import dot.codegenerator.modules.Mul_Generator;
import dot.codegenerator.modules.Nop_Generator;
import dot.codegenerator.modules.Number_Generator;
import dot.codegenerator.modules.Overload_Generator;
import dot.codegenerator.modules.PC_Generator;
import dot.codegenerator.modules.Put_Generator;
import dot.codegenerator.modules.Redefine_Generator;
import dot.codegenerator.modules.Ref_Generator;
import dot.codegenerator.modules.Return_Generator;
import dot.codegenerator.modules.Set_Generator;
import dot.codegenerator.modules.String_Generator;
import dot.codegenerator.modules.Sub_Generator;
import dot.codegenerator.modules.Symbol_Generator;
import dot.codegenerator.modules.Sync_Generator;
import dot.codegenerator.modules.UnRef_Generator;
import dot.codegenerator.modules.Undefine_Generator;
import dot.codegenerator.modules.UnwindList_Generator;
import dot.codegenerator.modules.Virtual_Generator;
import dot.codegenerator.modules.While_Generator;
import dot.codegenerator.optimizers.CreateRelativeJumpsProcessor;
import dot.codegenerator.optimizers.ForwardCallProcessor;
import dot.lexer.Constants;
import dot.lexer.Token;
import dot.parser.nodes.AtNode;
import dot.parser.nodes.BlockNode;
import dot.parser.nodes.INode;
import dot.parser.nodes.NumberNode;
import dot.parser.nodes.ProgramNode;
import dot.parser.nodes.StringNode;
import dot.parser.nodes.SymbolNode;
import dot.parser.nodes.TildeNode;

public class CodeGenerator implements ICodeGenerator {

	private INode rootNode;

	private ArrayList<IInstruction> programList;

	private Hashtable<String, IGenerator> generators;

	private Stack<Map<String, String>> environments;

	private Stack<ArrayList<String>> lambdaFrames;

	private Stack<String> loopBreakTargets;
	
	private Stack<String> loopContinueTargets;

	private Stack<String> functionBlocks;

	private Set<String> macronames;

	private List<Macro> macros;

	private boolean inLambda;

	public int jumpTargets = 0;

	public int tempVars = 0;

	public int xmlindent = 0;
	
	private String context = "";

	private ArrayList<String> userlabels;

	private Hashtable<String, String> redefinitions;

	public CodeGenerator(INode root) throws Exception {
		rootNode = reorderTree(root);
		generators = new Hashtable<String, IGenerator>();
		programList = new ArrayList<IInstruction>();
		environments = new Stack<Map<String, String>>();
		lambdaFrames = new Stack<ArrayList<String>>();
		inLambda = false;
		loopBreakTargets = new Stack<String>();
		loopContinueTargets = new Stack<String>();
		userlabels = new ArrayList<String>();
		functionBlocks = new Stack<String>();
		macronames = new HashSet<String>();
		
		macros = new ArrayList<Macro>();
		redefinitions = new Hashtable<String, String>();

		pushFunctionBlock(" ++ GLOBAL ++");
		pushEnvironment();

		/* Register built in code-generator modules */
		registerGenerator(new Call_Generator());
		registerGenerator(new Defun_Generator());
		generators.put("var", registerGenerator(new Defvar_Generator()));
		registerGenerator(new If_Generator());
		registerGenerator(new Number_Generator());
		registerGenerator(new String_Generator());
		registerGenerator(new Symbol_Generator());
		registerGenerator(new Begin_Generator());
		registerGenerator(new Let_Generator());
		registerGenerator(new Set_Generator());
		registerGenerator(new While_Generator());
		registerGenerator(new Cond_Generator());
		registerGenerator(new For_Generator());
		registerGenerator(new End_Generator());
		registerGenerator(new EndFun_Generator());
		registerGenerator(new Sync_Generator());
		registerGenerator(new Lambda_Generator());
		registerGenerator(new Break_Generator());
		registerGenerator(new Ref_Generator());
		registerGenerator(new UnRef_Generator());
		registerGenerator(new Virtual_Generator());
		registerGenerator(new Return_Generator());
		registerGenerator(new Goto_Generator());
		registerGenerator(new Label_Generator());
		registerGenerator(new Foreach_Generator());
		registerGenerator(new Equal_Generator());
		registerGenerator(new Get_Generator());
		registerGenerator(new Concat_Generator());
		registerGenerator(new Add_Generator());
		registerGenerator(new List_Generator());
		registerGenerator(new Dict_Generator());
		registerGenerator(new Append_Generator());
		registerGenerator(new Div_Generator());
		registerGenerator(new Mul_Generator());
		registerGenerator(new Put_Generator());
		registerGenerator(new Sub_Generator());
		registerGenerator(new Defmacro_Generator());
		registerGenerator(new Redefine_Generator());
		registerGenerator(new Undefine_Generator());
		registerGenerator(new Overload_Generator());
		registerGenerator(new Nop_Generator());
		registerGenerator(new PC_Generator());
		registerGenerator(new Continue_Generator());
		registerGenerator(new UnwindList_Generator());
		registerGenerator(new Context_Generator());

	}

	public INode expand(INode node) throws Exception {
		if (node instanceof BlockNode) {
			if (node.getChildCount() == 0)
				return node;
			INode expanded = node;
			String fname = expanded.getChild(0).getString();
			while (isMacro(fname)) {
				Macro m = getMacro(expanded);
				expanded = m.expand(this, expanded);
				fname = expanded.getChild(0).getString();
			}
			return expanded;
		}
		return node;
	}

	@SuppressWarnings("unchecked")
	public List<IInstruction> generate() throws Exception {
		scanTree(rootNode);
		ForwardCallProcessor cprocessor = new ForwardCallProcessor();
		programList = (ArrayList<IInstruction>) cprocessor.optimize(programList,
				this);
		CreateRelativeJumpsProcessor jprocessor = new CreateRelativeJumpsProcessor();
		programList = (ArrayList<IInstruction>) jprocessor.optimize(programList,
				this);
		return programList;
	}

	private boolean generateBlockNode(INode node) throws Exception {
		if (node.getChildCount() == 0)
			return false;

		String fname = node.getChild(0).getString();

		/* Handle special generators first */
		IGenerator gen = generators.get(fname);
		if (gen != null) {
			return gen.generate(programList, node);
			/* ...then handle definitions */
		} else if (fname.equals("define")) {
			if (node.getChild(1).getChildCount() > 0)
				return generators.get("defun").generate(programList, node);
			else
				return generators.get("defvar").generate(programList, node);
			/* ... handle lambda calls */
		} else if ((node.getChild(0).getChildCount() > 0)
				&& (node.getChild(0).getChild(0).getString().equals("lambda") || isVisibleVariable(node
						.getChild(0).getChild(0).getString()))) {
			/* Load parameters */
			for (int i = 1; i < node.getChildCount(); i++)
				generateCode(node.getChild(i));
			if (node.getChild(0).getChild(0).getString().equals("lambda")) {
				/* Inline lambda definition */
				IGenerator inlineLambdagen = generators.get(node.getChild(0)
						.getChild(0).getString());
				/* Generate lambda code */
				inlineLambdagen.generate(programList, node.getChild(0));
			} else {
				/*
				 * TODO: must load parameters on the stack and then call the
				 * lambda
				 */
				System.out.println("TODO");
			}
			/* Generate the call */
			Instruction fcall = null;
			fcall = new Instruction(IConstants.OPCODES.LCALL, node);
			fcall.pushParameter(node.getChildCount() - 1);
			programList.add(fcall);
			return false;
		} else if (isMacro(fname)) {
			Macro m = getMacro(node);
			generateCommonCode(m.expand(this, node));
			return false;
		} else {
			/* Handle normal function/procedure call */
			IGenerator genCall = generators.get("++ call ++");
			return genCall.generate(programList, node);
		}
	}

	public boolean generateCode(INode node) throws Exception {
		return generateCommonCode(node);
	}

	private boolean generateCommonCode(INode node) throws Exception {
		/*
		 * returns true if the scan routine should continue with children false
		 * if the children have been already processed
		 */
		if (node instanceof StringNode) {
			return generateStringNode(node);
		} else if (node instanceof NumberNode) {
			return generateNumberNode(node);
		} else if (node instanceof SymbolNode) {
			return generateSymbolNode(node);
		} else if (node instanceof ProgramNode) {
			return true;
		} else if (node instanceof BlockNode) {
			return generateBlockNode(node);
		} else if (node instanceof TildeNode) {
			return false;
		} else if (node instanceof AtNode) {
			return generateUnwindListNode(node);
		}
		System.out.println("Unknown node " + node.getClass());
		return false;
	}

	private boolean generateNumberNode(INode node) throws Exception {
		IGenerator gen = generators.get("++ number ++");
		return gen.generate(programList, node);
	}

	public String generateJumpTarget() {

		return "__target__ " + (++jumpTargets);
	}

	private boolean generateStringNode(INode node) throws Exception {
		IGenerator gen = generators.get("++ string ++");
		return gen.generate(programList, node);
	}

	private boolean generateSymbolNode(INode node) throws Exception {
		IGenerator gen = generators.get("++ symbol ++");
		return gen.generate(programList, node);
	}

	public String generateTempVarName() {

		return "var " + (++tempVars);
	}

	private boolean generateUnwindListNode(INode node) throws Exception {
		IGenerator gen = generators.get("++ unwindlist ++");
		return gen.generate(programList, node);
	}

	private INode getBogusProgramEndNode() {
		BlockNode b = new BlockNode(new Token(Constants.OPENBLOCK, "", 0, 0),
				"");
		SymbolNode end = new SymbolNode(new Token(Constants.SYMBOL,
				"++ end ++", "", 0, 0), "");
		b.addChild(end);
		return b;
	}

	public String getCurrentFunctionBlock() {
		return functionBlocks.peek();
	}

	public Object getEnvironment() {
		return environments.lastElement();
	}

	public String getFunctionName(String name) {
		if (redefinitions.containsKey(name)) {
			return redefinitions.get(name);
		} else
			return name;
	}

	/* ** Generic node type generators ** */

	/** Lambda vars * */

	public List<String> getLambdaVars() {
		ArrayList<String> vList = lambdaFrames.lastElement();
		return vList;
	}

	/** Loop break points * */

	public String getLoopBreakTarget() {
		if (loopBreakTargets.isEmpty())
			return null;
		return loopBreakTargets.lastElement();
	}

	public Macro getMacro(INode node) throws Exception {
		String sig = getMacroCallSignature(node);
		for (Macro m : macros) {
			if (m.matches(sig))
				return m;
		}
		Logger
				.CompilerFatal(node, this, "Error resolving macro for signature '%s'", sig);
		return null;
	}
	
	/* Loop continue points */
	
	public String getLoopContinueTarget() {
		if (loopContinueTargets.isEmpty())
			return null;
		return loopContinueTargets.lastElement();
	}

	public String popLoopContinueTarget() {
		if (loopContinueTargets.isEmpty())
			return null;
		return loopContinueTargets.pop();
	}

	public void pushLoopContinueTarget(String label) {
		loopContinueTargets.push(label);
	}


	public String getMacroCallSignature(INode node) throws Exception {
		StringBuilder sig = new StringBuilder();
		int argssize = node.getChildCount();
		for (int i = 0; i < argssize; i++) {
			INode child = node.getChild(i);
			String str = "none";
			if (child instanceof SymbolNode)
				str = child.getString();
			sig.append(str);
			sig.append(' ');
		}
		return sig.toString();
	}

	private boolean isFunctionCall(INode block) throws Exception {
		return (!isVarDefinition(block)) && (!isFunctionDefinition(block)) && (!isMacroDefinition(block));
	}

	private boolean isFunctionDefinition(INode block) throws Exception {
		if (block.getChildCount() < 1)
			return false;
		if (block.getChild(0).getString().equals("define")
				|| block.getChild(0).getString().equals("defun"))
			if (block.getChild(1).getChildCount() > 0)
				return true;
		return false;
	}

	public boolean isGlobalFunction(String fname) {
		return "f".equals(environments.firstElement().get(fname));
	}

	public boolean isGlobalVariable(String vname) {
		return "v".equals(environments.firstElement().get(vname));
	}

	public boolean isLocalFunction(String fname) {
		return "f".equals(environments.lastElement().get(fname));
	}

	public boolean isLocalVariable(String vname) {
		return "v".equals(environments.lastElement().get(vname));
	}

	public boolean isMacro(String name) {
		return macronames.contains(name);
	}

	public boolean isValidLabel(String label) {
		return !userlabels.contains(label);
	}

	private boolean isVarDefinition(INode block) throws Exception {
		if (block.getChildCount() < 2)
			return false;
		if (block.getChild(0).getString().equals("define")
				|| block.getChild(0).getString().equals("defvar")
				|| block.getChild(0).getString().equals("virtual"))
			if (block.getChild(1).getChildCount() == 0)
				return true;
		return false;
	}
	
	private boolean isMacroDefinition(INode block) throws Exception {
		if (block.getChildCount() < 3)
			return false;
		if (block.getChild(0).getString().equals("defmacro"))
				return true;
		return false;
	}

	public boolean isVisibleFunction(String vname) {
		for (int i = environments.size() - 1; i >= 0; i--) {
			Map<?, ?> env = environments.get(i);
			if ("f".equals(env.get(vname)))
				return true;
		}
		return false;
	}

	public boolean isVisibleVariable(String vname) {
		for (int i = environments.size() - 1; i >= 0; i--) {
			Map<?, ?> env = environments.get(i);
			if ("v".equals(env.get(vname)))
				return true;
		}
		return false;
	}

	public void overload(String genName, String newName) {
		if (generators.containsKey(genName)) {
			IGenerator gen = generators.get(genName);
			generators.put(newName, gen);
		} else {
			redefinitions.put(newName, genName);
		}
	}

	public void popEnvironment() {
		environments.pop();
	}

	public void popFunctionBlock() {
		functionBlocks.pop();
	}

	public void popLambdaFrame() {
		lambdaFrames.pop();
		if (lambdaFrames.size() < 1)
			inLambda = false;
	}

	public String popLoopBreakTarget() {
		if (loopBreakTargets.isEmpty())
			return null;
		return loopBreakTargets.pop();
	}

	/** Variable scope * */

	public void pushEnvironment() {
		environments.push(new HashMap<String, String>());
	}

	public void pushFunctionBlock(String fname) {
		functionBlocks.push(fname);
	}

	public void pushGlobalFunction(String fname) {
		environments.firstElement().put(fname, "f");
	}

	public void pushGlobalVariable(String vname) {
		environments.firstElement().put(vname, "v");
	}

	public void pushLabel(String label) {
		userlabels.add(label);
	}

	public void pushLambdaFrame() {
		lambdaFrames.push(new ArrayList<String>());
		inLambda = true;
	}

	public void pushLambdaVar(String name) {
		if (inLambda) {
			ArrayList<String> vList = (ArrayList<String>) lambdaFrames
					.lastElement();
			vList.add(name);
		}
	}

	public void pushLocalVariable(String vname) {
		environments.lastElement().put(vname, "v");
	}

	public void pushLoopBreakTarget(String label) {
		loopBreakTargets.push(label);
	}

	public void pushMacro(Macro macro) {
		String name = macro.getName();

		if (macros.contains(macro)) {
			System.err.format("Error: Redefinition of macro %s with signature %s\n", name, macro.getSignature());
			System.exit(1);
		} else {
			macros.add(macro);
			macronames.add(name);
			Collections.sort(macros);
		}
	}

	public void redefine(String genName, String newName) {
		if (generators.containsKey(genName)) {
			IGenerator gen = generators.remove(genName);
			generators.put(newName, gen);
		} else {
			redefinitions.put(newName, genName);
		}
	}

	public IGenerator registerGenerator(IGenerator g) {
		g.initialize(this);
		generators.put(g.getIdentifier(), g);
		return g;
	}

	public void registerPlugin(IGenerator plugin) {
		generators.put(plugin.getIdentifier(), plugin);
	}

	private INode reorderTree(INode root) throws Exception {
		INode newroot = new ProgramNode();
		for (int i = 0; i < root.getChildCount(); i++) {
			INode block = root.getChild(i);
			if (isMacroDefinition(block))
				newroot.addChild(block);
		}
		for (int i = 0; i < root.getChildCount(); i++) {
			INode block = root.getChild(i);
			if (isVarDefinition(block))
				newroot.addChild(block);
		}
		for (int i = 0; i < root.getChildCount(); i++) {
			INode block = root.getChild(i);
			if (isFunctionCall(block))
				newroot.addChild(block);
		}
		newroot.addChild(getBogusProgramEndNode());
		for (int i = 0; i < root.getChildCount(); i++) {
			INode block = root.getChild(i);
			if (isFunctionDefinition(block))
				newroot.addChild(block);
		}
		return newroot;
	}

	private void scanTree(INode node) throws Exception {
		if (generateCode(node))
			for (int i = 0; i < node.getChildCount(); i++)
				scanTree(node.getChild(i));
	}

	public void undefine(String genName) {
		if (generators.containsKey(genName)) {
			generators.remove(genName);
		} else {
			redefinitions.remove(genName);
		}
	}

	@Override
	public void setContext(String c) {
		context = c;
	}
	
	@Override
	public String getContext() {
		return context;
	}
	
	public void printMacroSignatures() {
		int t = 0;
		for (Macro m : macros) {
			System.out.format("MacroDescriptor temp%d = new MacroDescriptor(\"%s\", \"%s\");\n", t, m.getName(), m.getSignature());
			System.out.format("macros.put(temp%d.getPattern(), temp%d);\n", t, t);
			t++;
		}
	}

}
