/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package dot.codegenerator.macros;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import logging.Logger;
import dot.codegenerator.ICodeGenerator;
import dot.parser.nodes.AtNode;
import dot.parser.nodes.BackSlashNode;
import dot.parser.nodes.ChapNode;
import dot.parser.nodes.INode;
import dot.parser.nodes.SymbolNode;

public class Macro implements Comparable<Macro> {
	
	private String name = null;
	
	private List<String> arguments;
	
	private INode template;
	
	private String sig;
	
	private String salt;
	
	private Pattern pattern;
	
	// By default macro arguments are generated to be unique
	// by adding a ! in front of an argument it is possible
	// to force the name (without the !)
	// @ is for variable number of parameters 
	// !@ is valid
	
	public Macro(ICodeGenerator cg, INode node) throws Exception {
		name = node.getChild(1).getChild(0).getString();
		StringBuilder signature = new StringBuilder();
		signature.append(name);
		arguments = new ArrayList<String>();
		int argssize = node.getChild(1).getChildCount();
		for (int i=1; i < argssize;) {
			INode child = node.getChild(1).getChild(i);
			if (child instanceof BackSlashNode) {
				String fixedparam = '\\'+node.getChild(1).getChild(++i).getString();
				arguments.add(fixedparam);
				signature.append("\\s+");
				signature.append(fixedparam.substring(1));
			} else if (child instanceof ChapNode) {
				
				String avoidparam ="(\\s+[^("+node.getChild(1).getChild(++i).getString()+")])*";
				signature.append("\\s+");
				signature.append(avoidparam);
			} else if (child instanceof AtNode) {
				String varparam = '@'+node.getChild(1).getChild(++i).getString();
				if ((i != argssize - 1) && !(node.getChild(1).getChild(i+1) instanceof BackSlashNode)
						&& !(node.getChild(1).getChild(i+1) instanceof ChapNode)) {
					Logger.CompilerFatal(node, cg, "Variable argument not followed by fixed tag in macro definition");
				}
				arguments.add(varparam);
				signature.append("(\\s+[\\w.$/:\\-\\!~]+)+");
			} else {
				signature.append("\\s+[\\w.$/:\\-\\!~]+");
				arguments.add(child.getString());
			}
			i++;
		}
		signature.append("\\s*");
		pattern = Pattern.compile(signature.toString());
		template = node.getChild(2).clone();
		salt = String.format("%d",template.hashCode());
		makeUniqueSymbols(template);
		sig = signature.toString();
	}
	
	public String getSignature() {
		return sig.toString();
	}
	
	public boolean matches(String sig) {
		Matcher matcher = pattern.matcher(sig);
		return matcher.matches();	
	}
	
	public void pushArgument(String name) {
		arguments.add(name);
	}
	
	public String getName() {
		return name;
	}
	
	public int hashCode() {
		return sig.hashCode();
	}
	
	public boolean equals(Object o) {
		if (!(o instanceof Macro)) return false;
		Macro o2 = (Macro) o;
		return sig.equals(o2.getSignature());
	}
	
	public INode expand(ICodeGenerator cg, INode block) throws Exception {		
		INode newblock = template.clone(); // Create a blank copy
		Map<String,Object> repl = new HashMap<String,Object>();
		for (int i=1, j=0; i < block.getChildCount() &&  j < arguments.size();) {
			
			if (arguments.get(j).charAt(0) == '@') {
				/** ***************************************
				 *  Variable argument 
				 *  *************************************** */
				
				String next = null;
				if (j+1 < arguments.size())
					next = arguments.get(j+1).substring(1);
				
				List<INode> list = new ArrayList<INode>();
				
				for (;i < block.getChildCount();) {
					
					if (block.getChild(i).getString().equals(next)) {
						j++;
						break;
					}
					
					list.add(block.getChild(i));
					
					i++;
					
					if (i == block.getChildCount()) { // End of block
						if (next != null) {
							Logger.CompilerFatal(block, cg, "Unexpected end of macro block, expecting '%s'", next);
						}
						j++;
						break;
					}
				}
				repl.put(arguments.get(j-1).substring(1), list);
	
			} else if (arguments.get(j).charAt(0) != '\\') {
				
				repl.put(arguments.get(j), block.getChild(i));
				j++;
				i++;

			} else {
	
				if (!arguments.get(j).equals('\\'+block.getChild(i).getString())) {
					Logger.CompilerFatal(block, cg, "Failed macro substitution: expecting '%s' but got '\\%s'", arguments.get(j),block.getChild(i).getString());
					System.exit(1);
				}
				j++;
				i++;
				
			}
		}
		traverseAndReplace(newblock, repl);
		return newblock;
	}
	
	@SuppressWarnings("unchecked")
	private void traverseAndReplace(INode root, Map<String,Object> replacements) throws Exception {
		for (int i=0; i < root.getChildCount(); i++) {
			INode c = root.getChild(i);
			if (c instanceof SymbolNode) {
				String symbol = c.getString();
				if (replacements.containsKey(symbol)) {
					if (replacements.get(symbol) instanceof List) {
						List<INode> block = (List<INode>) replacements.get(symbol);
						root.removeChild(i);
						for (int k= block.size() -1; k >= 0; k--) {
							INode rchild = block.get(k);
							root.addChildAt(i, rchild.clone());
						}
					} else {
						root.replaceChild(i, ((INode) replacements.get(symbol)).clone());
					}
				}
			} else if (c.getChildCount() > 0) {
				traverseAndReplace(c, replacements);
			}
		}	
	}
	
	private void makeUniqueSymbols(INode root) throws Exception {
		for (int i=0; i < root.getChildCount(); i++) {
			INode c = root.getChild(i);
			if (c instanceof SymbolNode) {
				String symbol = c.getString();
				//if (arguments.contains(symbol)) continue; // Do not replace arguments
				String newSymbol = null;
				if (symbol.endsWith("#")) {
					newSymbol = String.format("%s_%s", symbol.substring(0,symbol.length()-2), salt);
					c.setString(newSymbol); // Generate unique symbol for the rest
					if (arguments.contains(symbol)) {
						int index = arguments.indexOf(symbol);
						arguments.remove(index);
						arguments.add(index, newSymbol);
					} else if (arguments.contains("@"+symbol)) {
						int index = arguments.indexOf("@"+symbol);
						arguments.remove(index);
						arguments.add(index, "@"+newSymbol);
					}
				}
			} else if (c.getChildCount() > 0) {
				makeUniqueSymbols(c);
			}
		}	
	}

	@Override
	public int compareTo(Macro o2) {
		Integer tl = new Integer(sig.length());
		Integer o2l = new Integer(o2.getSignature().length());
		return o2l.compareTo(tl);
	}

}