/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package dot.codegenerator.modules;


import java.util.List;

import com.fuzzbug.solenopsis.sdk.lib.IConstants;
import com.fuzzbug.solenopsis.sdk.lib.IInstruction;


import logging.Logger;

import dot.codegenerator.ICodeGenerator;
import dot.codegenerator.Instruction;
import dot.parser.nodes.BlockNode;
import dot.parser.nodes.INode;

public class Foreach_Generator implements IGenerator {

	ICodeGenerator cg;

	@SuppressWarnings("unchecked")
	public boolean generate(List programList, INode node) throws Exception {
		String foreachExitTag = cg.generateJumpTarget();
		String foreachStartTag = cg.generateJumpTarget();
		String tempVariable = cg.generateTempVarName();
		String list = cg.generateTempVarName();

		String varname = node.getChild(1).getString();
		String intag = node.getChild(2).getString();

		if (!intag.equals("in")) {
			Logger.CompilerFatal(node, cg, "Expecting 'in' tag in foreach statement");
		}
		
		INode listBlock = node.getChild(3);
		INode body = node.getChild(4);
	
		cg.pushLoopBreakTarget(foreachExitTag);
		cg.pushLoopContinueTarget(foreachStartTag);
		cg.pushEnvironment();
		/* Create local for environment */
		IInstruction local;
		local = new Instruction(IConstants.OPCODES.LOCAL, node);
		programList.add(local);
		/* Generate list and save in list variable */
		cg.generateCode(listBlock);

		/* Store in a temp variable (the for index) */
		Instruction storelist = new Instruction(IConstants.OPCODES.STORE, node);
		storelist.pushParameter(list);
		programList.add(storelist);	
		cg.pushLocalVariable(list);
		/* Load 0 on the stack */
		Instruction constOp = new Instruction(IConstants.OPCODES.CONST, node);
		constOp.pushParameter(0);
		programList.add(constOp);
		/* Store in a temp variable (the for index) */
		Instruction store = new Instruction(IConstants.OPCODES.STORE, node);
		store.pushParameter(tempVariable);
		programList.add(store);
		cg.pushLocalVariable(tempVariable);
		/* Load nil on the stack */
		Instruction loadnil = new Instruction(IConstants.OPCODES.CONST, node);
		loadnil.pushNil();
		programList.add(loadnil);
		/* Store in the var variable */
		Instruction storevar = new Instruction(IConstants.OPCODES.STORE, node);
		storevar.pushParameter(varname);
		programList.add(storevar);
		cg.pushLocalVariable(varname);
		/* start label */
		Instruction sLabel = new Instruction(IConstants.OPCODES.LABEL, node);
		sLabel.pushParameter(foreachStartTag);
		programList.add(sLabel);
		/* Load the temp variable (the for index) */
		Instruction load = new Instruction(IConstants.OPCODES.LOAD, node);
		load.pushParameter(tempVariable);
		programList.add(load);
		/* Load the list */
		Instruction load2 = new Instruction(IConstants.OPCODES.LOAD, node);
		load2.pushParameter(list);
		programList.add(load2);
		/* Call len */
		IInstruction lencall = new Instruction(IConstants.OPCODES.LEN, node);
		programList.add(lencall);
		/* Do test */
		IInstruction testcall = new Instruction(IConstants.OPCODES.LT, node);
		programList.add(testcall);
		/* Test the condition */
		Instruction fJump = new Instruction(IConstants.OPCODES.FJUMP, node);
		fJump.pushParameter(foreachExitTag);
		programList.add(fJump);
		/* Load the list */
		Instruction load3 = new Instruction(IConstants.OPCODES.LOAD, node);
		load3.pushParameter(list);
		programList.add(load3);
		/* Load the temp variable (the for index) */
		Instruction load4 = new Instruction(IConstants.OPCODES.LOAD, node);
		load4.pushParameter(tempVariable);
		programList.add(load4);
		/* Get item */
		IInstruction itemcall = new Instruction(IConstants.OPCODES.ITEM, node);
		programList.add(itemcall);
		/* Store the item in varname */
		Instruction set = new Instruction(IConstants.OPCODES.SET, node);
		set.pushParameter(varname);
		programList.add(set);
		/* Load the temp variable (the for index) */
		Instruction load5 = new Instruction(IConstants.OPCODES.LOAD, node);
		load5.pushParameter(tempVariable);
		programList.add(load5);
		/* Increments hte value */
		IInstruction sumcall = new Instruction(IConstants.OPCODES.INC, node);
		programList.add(sumcall);
		/* Store back in temp variable */
		Instruction set2 = new Instruction(IConstants.OPCODES.SET, node);
		set2.pushParameter(tempVariable);
		programList.add(set2);
		/* Generate body */
		cg.generateCode(body);
		/* jump to start */
		Instruction lJump = new Instruction(IConstants.OPCODES.JUMP, node);
		lJump.pushParameter(foreachStartTag);
		programList.add(lJump);
		/* exit label */
		Instruction eLabel = new Instruction(IConstants.OPCODES.LABEL, node);
		eLabel.pushParameter(foreachExitTag);
		programList.add(eLabel);
		/* Drop the local environment */
		IInstruction dropOp = new Instruction(IConstants.OPCODES.DROP, node);
		programList.add(dropOp);
		/* Push something on the stack if the body was empty */
		if ((body instanceof BlockNode) && (body.getChildCount() == 0)) {
			Instruction nil = new Instruction(IConstants.OPCODES.CONST, node);
			nil.pushNil();
			programList.add(nil);
		}

		cg.popLoopBreakTarget();
		cg.popLoopContinueTarget();
		cg.popEnvironment();
		return false;
	}

	public String getIdentifier() {
		return "foreach";
	}

	public void initialize(ICodeGenerator codegen) {
		cg = codegen;
	}

}
