/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package dot.lexer;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.List;

import logging.Logger;



public class Lexer {

	private char ch;

	private BufferedReader source;

	private boolean skipNextRead = false;

	private int line = 0;

	private int pos = 0;

	private boolean inComment = false;
	
	private int offset = 0;
	
	private List<String> lines;
	
	private String filename;

	public Lexer(String filename, List<String> code, int offset) throws IOException {
		ByteArrayOutputStream os = new ByteArrayOutputStream();	
		for (String l : code) {
			os.write(l.getBytes());
		}
		os.close();
		this.offset = offset;
		line = -offset + 1;
		lines = code;
		this.filename = filename;
		source = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(os.toByteArray())));
	}
	
	public int getLine() {
		return line;
	}
	
	public String getFilename() {
		return filename;
	}
	
	public String getLineContents(int lnumber) {
		int index = offset + lnumber - 1;
		return lines.get(index);
	}
	
	public int getPosition() {
		return pos - 1;
	}

	public Token getToken() throws Exception {
		readChar();
		consumeWhitespaces();
		switch(ch) {
		case (char) -1:
			return new Token(Constants.EOF, filename, line, pos);
		case '(':
			return new Token(Constants.OPENBLOCK, filename, line, pos);
		case ')':
			return new Token(Constants.CLOSEBLOCK, filename, line, pos);
		case '{':
			return new Token(Constants.OPENDICTBLOCK, filename, line, pos);
		case '}':
			return new Token(Constants.CLOSEDICTBLOCK, filename, line, pos);
		case '[':
			return new Token(Constants.OPENLISTBLOCK, filename, line, pos);
		case ']':
			return new Token(Constants.CLOSELISTBLOCK, filename, line, pos);			
		case '\'':
			return new Token(Constants.QUOTE, filename, line, pos);
		case '`':
			return new Token(Constants.BACKQUOTE, filename, line, pos);
		case ':':
			return new Token(Constants.SEMICOLON, filename, line, pos);			
		case ',':
			return new Token(Constants.COMMA, filename, line, pos);
		case '@':
			return new Token(Constants.AT, filename, line, pos);
		case '^':
			return new Token(Constants.CHAP, filename, line, pos);
		case '.':
			return new Token(Constants.DOT, filename, line, pos);
		case '\\':
			return new Token(Constants.BACKSLASH, filename, line, pos);
		case '°':
			return new Token(Constants.CIRCLE, filename, line, pos);
		case '~':
			return new Token(Constants.CHECK, filename, line, pos);
		case '"':
			return readString();
		}
		// Handle special 
		if (Character.isDigit(ch))
			return readNumber((char) -1);
		else if (ch == '-') {
			readChar();
			skipNextReadChar();
			if (Character.isDigit(ch))
				return readNumber('-');
			else
				return readSymbol('-');
		} else if (ch == '/') {
			readChar();
			skipNextReadChar();
			if (ch == '*') {
				consumeLongComment();
				return getToken();
			} else
				return readSymbol('/');
		} else
			return readSymbol((char) -1);
	}
	
	private void readChar() throws IOException {
		if (!skipNextRead) {
			ch = (char) source.read();
			checkNewLine();
		}
		skipNextRead = false;
	}

	private void skipNextReadChar() throws IOException {
		skipNextRead = true;
	}

	private void consumeWhitespaces() throws IOException {
		while (Character.isWhitespace(ch) || (ch <= 0) || inComment
				|| (ch == ';')) {
			readChar();
			if (ch == ';')
				inComment = true;	
		}
	}
	
	private void consumeLongComment() throws IOException {
		char prev = ' ';
		ch = ' ';
		while (!((ch == '/') && (prev == '*'))) {
			prev = ch;
			readChar();
		}
	}

	private void checkNewLine() {
		if (ch == '\n') {
			inComment = false;
			line++;
			pos = 0;
		} else
			pos++;
	}
	
	public List<String> getLines() {
		return lines;
	}

	private Token readString() throws Exception {
		int line = this.line; // Save the line
		int pos = this.pos; // Save the position
		String tmpString = new String();
		readChar();
		while ((ch != '"') && (ch != (char) -1)) {
			if (ch == '\n') {
				Logger.fatal("Error! Unterminated string at line %d, position %d\n", line, pos);
			}
			tmpString += ch;
			readChar();
		}
		return new Token(Constants.STRING, tmpString, filename, line, pos);
	}

	private boolean isValidSymbolCharacter(char c) {
        if ((c == '_') || (c == '$') || (c == '!') || (c == '?') || (c == '*') || (c == '%')
                        || (c == '+') || (c == '-') || (c == '=') || (c == ':')
                        || (c == '<') || (c == '>') || (c == '/') || (c == '#') )
                return true;
        else
                return Character.isLetterOrDigit(c);
	}

	private Token readSymbol(char prev) throws IOException {
		int line = this.line; // Save the line
		int pos = this.pos; // Save the position
		String tmpSymbol = new String();
		if (prev != (char) -1)
			tmpSymbol += prev;
		while (isValidSymbolCharacter(ch)) {
			tmpSymbol += ch;
			readChar();
		}
		skipNextReadChar();
		return new Token(Constants.SYMBOL, tmpSymbol, filename, line, pos);
	}

	private Token readNumber(char prev) throws IOException {
		int line = this.line; // Save the line
		int pos = this.pos; // Save the position
		String num = new String();
		int sign = 1;
		if (prev == '-') {
			sign = -1;
			readChar();
		}
		while (Character.isDigit(ch)) {
			num += ch;
			readChar();
		}
		if (ch == 'l') {
			long intValue = new Long(num) * sign;
			return new Token(Constants.LONGNUMBER, intValue, filename, line, pos);
		} else if (ch == '.') {
			num = num + ch;
			readChar();
			while (Character.isDigit(ch)) {
				num = num + ch;
				readChar();
			}
			
			if (ch == 'f') {
				float floatValue = new Float(num).floatValue() * sign;
				return new Token(Constants.FLOATNUMBER, floatValue, filename, line, pos);
			} else {
				skipNextReadChar();
				double doubleValue = new Double(num).doubleValue() * sign;
				return new Token(Constants.DOUBLENUMBER, doubleValue, filename, line, pos);
			}
		}
		int intValue = new Integer(num).intValue() * sign;
		skipNextReadChar();
		return new Token(Constants.INTNUMBER, intValue, filename, line, pos);

	}

}
