/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package dot.parser;

import java.io.IOException;

import logging.Logger;

import dot.lexer.Constants;
import dot.lexer.Lexer;
import dot.lexer.Token;
import dot.parser.nodes.AtNode;
import dot.parser.nodes.BackQuoteNode;
import dot.parser.nodes.BackSlashNode;
import dot.parser.nodes.BlockNode;
import dot.parser.nodes.ChapNode;
import dot.parser.nodes.CommaNode;
import dot.parser.nodes.DotNode;
import dot.parser.nodes.INode;
import dot.parser.nodes.NumberNode;
import dot.parser.nodes.ProgramNode;
import dot.parser.nodes.QuoteNode;
import dot.parser.nodes.SemicolonNode;
import dot.parser.nodes.StringNode;
import dot.parser.nodes.SymbolNode;
import dot.parser.nodes.TildeNode;



public class Parser {

	private Lexer lex;

	private Token tok;
	
	private int blockOpen = 0;

	public Parser(Lexer lexer) throws IOException {
		lex = lexer;
	}

	public INode getParseTree() throws Exception {
		ProgramNode root = new ProgramNode();
		INode node = null;
		do {
			node = parse();
			if (!(node == null)) {
				root.addChild(node);
			}
		} while (!(node == null));
		return root;
	}

	private void getToken() throws Exception {
		tok = lex.getToken();
	}

	private INode parse() throws Exception {
		getToken();
		switch (tok.getType()) {
		case Constants.OPENBLOCK:
			blockOpen++;
			return parseBlock();
		case Constants.OPENDICTBLOCK:
			blockOpen++;
			return parseDictBlock();
		case Constants.OPENLISTBLOCK:
			blockOpen++;
			return parseListBlock();
		case Constants.INTNUMBER:
			return new NumberNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.FLOATNUMBER:
			return new NumberNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.LONGNUMBER:
			return new NumberNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.DOUBLENUMBER:
			return new NumberNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.STRING:
			return new StringNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.SYMBOL:
			return new SymbolNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.COMMA:
			return new CommaNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.AT:
			return new AtNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.SEMICOLON:
			return new SemicolonNode(tok, lex.getLineContents(tok.getLine()));			
		case Constants.DOT:
			return new DotNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.BACKQUOTE:
			return new BackQuoteNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.CHAP:
			return new ChapNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.BACKSLASH:
			return new BackSlashNode(tok, lex.getLineContents(tok.getLine()));			
		case Constants.QUOTE: /* This need to be fixed */
			return new QuoteNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.CHECK:
			if (blockOpen != 0) {
				Logger.fatal(String.format("Parenthesis mismatch @ in file %s line %d, position %d: %s", lex.getFilename(), lex.getLine(), lex.getPosition(), lex.getLineContents(lex.getLine())));
			}
			return new TildeNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.CIRCLE:
			return new TildeNode(tok, lex.getLineContents(tok.getLine()));
		case Constants.EOF:
			if (blockOpen != 0)
				Logger.fatal(String.format("Unexpected end of file @ in file %s line %d, position %d", lex.getFilename(), lex.getLine(), lex.getPosition()));
			return null;
		case Constants.CLOSEBLOCK:
		case Constants.CLOSEDICTBLOCK:
		case Constants.CLOSELISTBLOCK:
			blockOpen--;
			if (blockOpen < 0) {
				Logger.fatal(String.format("Parenthesis mismatch @ in file %s line %d, position %d: %s", lex.getFilename(), lex.getLine(), lex.getPosition(), lex.getLineContents(lex.getLine())));
			}
			return null;
		default:
			throw new Exception(String.format("Unknown token type %d @ in file %s line %d, position %d: %s", tok.getType(), lex.getFilename(), lex.getLine(), lex.getPosition(), lex.getLineContents(lex.getLine())));
		}
	}

	private INode parseBlock() throws Exception {
		BlockNode block = new BlockNode(tok, lex.getLineContents(tok.getLine()));
		INode node;
		do {
			node = parse();
			if (!(node == null)) {
				block.addChild(node);
			}
		} while (!(node == null));
		return block;
	}
	
	private INode parseDictBlock() throws Exception {
		BlockNode keysblock = new BlockNode(new Token(Constants.OPENBLOCK,
				tok.getFilename(), tok.getLine(), tok.getPosition()), lex.getLineContents(tok.getLine()));
		SymbolNode keysblocklist = new SymbolNode(new Token(Constants.SYMBOL,
				"list", tok.getFilename(), tok.getLine(), tok.getPosition()), lex.getLineContents(tok.getLine()));
		keysblock.addChild(keysblocklist);
		
		BlockNode valuesblock = new BlockNode(new Token(Constants.OPENBLOCK,
				tok.getFilename(), tok.getLine(), tok.getPosition()), lex.getLineContents(tok.getLine()));
		SymbolNode valuesblocklist = new SymbolNode(new Token(Constants.SYMBOL,
				"list", tok.getFilename(), tok.getLine(), tok.getPosition()), lex.getLineContents(tok.getLine()));
		valuesblock.addChild(valuesblocklist);
		
		BlockNode dictblock = new BlockNode(new Token(Constants.OPENBLOCK,
				tok.getFilename(), tok.getLine(), tok.getPosition()), lex.getLineContents(tok.getLine()));
		SymbolNode dict = new SymbolNode(new Token(Constants.SYMBOL,
				"dict", tok.getFilename(), tok.getLine(), tok.getPosition()), lex.getLineContents(tok.getLine()));
		dictblock.addChild(dict);
		
		
		INode node;
		do {
			node = parse(); // We expect a string key
			
			if (node == null) break;
			
			if (!(node == null) && !(node instanceof SemicolonNode) && (node instanceof StringNode)) {
				keysblock.addChild(node);
			} else {
				Logger.fatal("Malformed dictionary: expecting string key in file %s at line %d, position %d", tok.getFilename(), tok.getLine(), tok.getPosition());
			}
			node = parse();
			if ((node == null) || !(node instanceof SemicolonNode)) {
				Logger.fatal("Malformed dictionary: expecting semicolon in file %s at line %d, position %d", tok.getFilename(), tok.getLine(), tok.getPosition());
			}
			node = parse();
			if (!(node == null) && !(node instanceof SemicolonNode)) {
				valuesblock.addChild(node);
			} else {
				Logger.fatal("Malformed dictionary: expecting value in file %s at line %d, position %d", tok.getFilename(), tok.getLine(), tok.getPosition());
			}
			node = parse();
			if ((node != null) && !(node instanceof CommaNode)) {
				Logger.fatal("Malformed dictionary: expecting comma in file %s at line %d, position %d", tok.getFilename(), tok.getLine(), tok.getPosition());
			}
		} while (!(node == null));
		
		if (keysblock.getChildCount() != valuesblock.getChildCount()) {
			Logger.fatal("Malformed dictionary in file %s at line %d, position %d", tok.getFilename(), tok.getLine(), tok.getPosition());
		}
		
		dictblock.addChild(keysblock);
		dictblock.addChild(valuesblock);
		return dictblock;
	}
	
	private INode parseListBlock() throws Exception {
		BlockNode listblock = new BlockNode(new Token(Constants.OPENBLOCK,
				tok.getFilename(), tok.getLine(), tok.getPosition()), lex.getLineContents(lex.getLine()));
		SymbolNode list = new SymbolNode(new Token(Constants.SYMBOL,
				"list", tok.getFilename(),  tok.getLine(), tok.getPosition()), lex.getLineContents(lex.getLine()));
		listblock.addChild(list);
		INode node;
		do {
			node = parse();
			if (!(node == null)) {
				listblock.addChild(node);
			} else break;
			node = parse();
			if ((node != null) && !(node instanceof CommaNode)) {
				Logger.fatal("Malformed list in file %s at line %d, position %d", tok.getFilename(), tok.getLine(), tok.getPosition());
			}
		} while (!(node == null));
		return listblock;
	}

}
