/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package exe;

import interfaces.exe.ICodeProvider;
import interfaces.softthreading.IScheduler;
import interfaces.vm.util.ISerializer;

import java.io.ByteArrayOutputStream;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;

import logging.Logger;
import scheduling.NodeSchedulingService;
import serialization.Serializer;
import softthreading.Scheduler;
import softthreading.SimpleSoftThread;
import bebop.client.Client;
import bebop.lib.Message;

import com.fuzzbug.solenopsis.sdk.exe.INodeDaemon;
import com.fuzzbug.solenopsis.sdk.plugins.IPlugin;
import com.fuzzbug.solenopsis.sdk.scheduling.INodeAlarmService;
import com.fuzzbug.solenopsis.sdk.vm.IVM;
import com.fuzzbug.solenopsis.sdk.vm.IVMState;

import core.signals.SignalEmitter;
import core.signals.signals.ExternCallSignalData;
import core.vm.lib.VMBaseOperations;
import core.vm.lib.VMCode;
import core.vm.lib.VMStack;
import core.vm.lib.VMState;
import dot.codegenerator.Instruction;
import dot.parser.nodes.INode;


public class NodeDaemon extends SimpleSoftThread implements INodeDaemon {
	
	private IScheduler 							scheduler = new Scheduler();
	private String 								identifier;
	private Map<String,IPlugin>					plugins = new HashMap<String,IPlugin>();
	private Hashtable<String,Object[]>			pluginMethodCache = new Hashtable<String,Object[]>();
	private Hashtable<String,Object>			properties = new Hashtable<String,Object>();
	private	ICodeProvider 						cp; 
	
	
	public NodeDaemon(String identifier, ICodeProvider c) {
		this.identifier = identifier;
		this.cp = c;
		registerPlugin(VMBaseOperations.instance());
	}

	@Override
	public String getIdentifier() {
		return identifier;
	}

	@Override
	public int getScheduleLength() {
		return scheduler.count();
	}

	@Override
	public void schedule(IVM vm) {
		VMStats.start(vm.getCode().getSourcePath());
		scheduler.schedule(vm);
	}

	@Override
	public boolean runextended() {
		return scheduler.execute();
	}

	@Override
	public void unschedule(IVM vm) {
		vm.stop();
		scheduler.unschedule(vm);
	}
	
	private String getNamespace (String name) {
		int index = name.indexOf(IPlugin.NAMESPACE_SEPARATOR);
		if (index > 0) {
			return name.substring(0, index);
		} else {
			return "";
		}
	}
	
	private String getTag (String name) {
		int index = name.indexOf(IPlugin.NAMESPACE_SEPARATOR);
		if (index > 0) {
			return name.substring(index+2);
		} else {
			return name;
		}
	}
	
	private Method resolveMethod (IPlugin p, String name) {
		String tag = getTag(name);
		if (p != null) {
			return p.getMethod(tag);
		}
		return null;
	}
	
	private IPlugin resolvePlugin (String name) {
		String ns = getNamespace(name);
		return plugins.get(ns);
	}

	public boolean hasMethod(String name) {
		return resolveMethod(resolvePlugin(name), name) != null;
	}

	public boolean registerPlugin(IPlugin p) {
		if(plugins.containsKey(p.getNamespaceName())) return false;
		try {
			p.doInitialize(this);
			plugins.put(p.getNamespaceName(), p);
		} catch (Exception e) {
			Logger.warning("Failed to initialize plugin '%s' : %s\n", p, e.getLocalizedMessage());
			return false;
		}
		return true;
	}

	public void unregisterPlugin(String ns) {
		cleanupPlugin(plugins.remove(ns));
	}
	
	private void cleanupPlugin(IPlugin p) {
		assert(p != null);
		List<String> toRemove = new ArrayList<String>();
		for (String o : pluginMethodCache.keySet()) {
			String nspace = getNamespace(o);
			if (nspace.equals(p.getNamespaceName()))
				toRemove.add(o);
		}
		for (String key : toRemove)
			pluginMethodCache.remove(key);
	}
	
	/**
	 * Requests synchronization to the synchronization provider.
	 * This method is called by the VM.
	 * @param source
	 * @param data
	 */
	public void vmSyncRequest (SignalEmitter source, Object data) {
		IVM sourceVM = (IVM) source;
		sourceVM.suspend();
		sourceVM.setPC(sourceVM.getPC()+1);
	}
	
	/**
	 * Requests stopping of the VM.
	 * This method is called by the VM.
	 * @param source
	 * @param data
	 */
	public void vmStopRequest (SignalEmitter source, Object data) {
		IVM sourceVM = (IVM) source;
		VMStats.stop(sourceVM.getCode().getSourcePath());
		sourceVM.stop();
	}

	@SuppressWarnings("unchecked")
	public Object vmExternOpCall (SignalEmitter source, Object data) throws Exception {
		ExternCallSignalData fdata = (ExternCallSignalData) data;
		String name = fdata.getName();
		Method m = null;
		IPlugin p = null;
		if (pluginMethodCache.containsKey(name)) {
			Object[] o = pluginMethodCache.get(name);
			p = (IPlugin) o[0];
			m = (Method) o[1];
		} else {
			p = resolvePlugin(name);
			m = resolveMethod(p, name);
			if ((p != null) && (m != null)) {
				Object[] o = new Object[2];
				o[0] = p;
				o[1] = m;
				pluginMethodCache.put(name, o);
			}
		}
		Object result = null;
		if (m != null) {
			try {
				result = m.invoke(p, (INodeDaemon) this, (IVM) source, (List) fdata.getArgs());
			} catch (Exception e) {
				Logger.warning("Failed call to '%s'\n", fdata.getName());
				pluginMethodCache.remove(name);
				e.printStackTrace();
			}
		} else {
			IVM vm = (IVM) source;
			Instruction instruction = (Instruction) vm.getCode().getOp(vm.getPC());
			INode inode = instruction.getNode();
			try {
				Logger.fatal("Error: external operator '%s' not bound in '%s' @ line: %d, pos: %d : %s\n", fdata.getName(), vm.getCode().getSourcePath(), inode.getLine(), inode.getPosition(), inode.getCode());
			} catch (Exception e) {
				Logger.fatal("Error: external operator '%s' not bound in '%s'\n", fdata.getName(), vm.getCode().getSourcePath());
			}
		}
		if (result == null) {
			Logger.warning("External operator '%s' returned null (converted to false)\n", fdata.getName());
			result = false;
		}
		return result;
	}

	public void setIdentifier(String identifier) {
		this.identifier = identifier;		
	}

	@Override
	public void doFinalize() {
		for (IPlugin p : plugins.values()) {
			try {
				p.doFinalize();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	@Override
	public Set<String> getNamespaces() {
		return plugins.keySet();
	}

	@Override
	public IPlugin getPlugin(String namespace) {
		return plugins.get(namespace);
	}
	
	public IVMState getState(String species) {
		IVMState state = null;
		try {
			state = new VMState(new VMStack(), new VMCode(cp, species), 0);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return state;
	}
	
	public void sendState(String recipient, IVMState s) {
		try {
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			ISerializer ser = new Serializer(out);
			try {
				ser.print(s.toBaseValue());
			} catch (Exception e) {
				e.printStackTrace();
			}
			out.flush();
			String state = out.toString();
			out.close();
			
			Map<String,Object> contents = new HashMap<String, Object>();
			contents.put("type", "ant");
			contents.put("source", this.getIdentifier());
			contents.put("state", state);
			contents.put("pc", 0);
			
			Message msg = new Message(recipient, contents);
			
			Client.sendMessage(msg);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public INodeAlarmService getNodeAlarmService() {
		return NodeSchedulingService.instance();
	}

	@Override
	public Object getProperty(String key) {
		return properties.get(key);
	}

	@Override
	public void setProperty(String key, Object value) {
		properties.put(key, value);
	}
	
	
}