/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package logging;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

public class StaticLogger {

	private static ArrayList<PrintStream> outputters;

	private static HashMap<PrintStream, Integer> levels;

	private static int defaultLevel = 3;
	
	private static boolean detailedOutput = false;

	@SuppressWarnings("unchecked")
	public static void log(int level, Object obj, String Message) {
		if(detailedOutput) {
			StackTraceElement[] st = Thread.currentThread().getStackTrace();
			Class klass = obj.getClass();
			for (int i = st.length - 1; i >= 0; i--) {
				String className = st[i].getClassName();
				String methodName = st[i].getMethodName();
				if (className.equals(klass.getName()) && !methodName.equals("main")) {
					int lineNumber = st[i].getLineNumber();
					outputMessage(level, String.format("<!%s:%d> %s \n", className,
							lineNumber, Message));
					return;
				}
			}
			String className = st[st.length - 1].getClassName();
			int lineNumber = st[st.length - 1].getLineNumber();
			outputMessage(level, String.format("<?%s:%d> %s \n", className,
					lineNumber, Message));
		} else {
			outputMessage(level, String.format("<??:??> %s \n", Message));
		}
	}

	public static void log(int level, Object obj, String Message, Object... objects) {
		log(level, obj, String.format(Message, objects));
	}

	public static void log(Object obj, String Message) {
		log(defaultLevel, obj, Message);
	}

	public static void log(Object obj, String Message, Object... objects) {
		log(defaultLevel, obj, String.format(Message, objects));
	}

	public static void log(String Message) {
		log(defaultLevel, Message);
	}

	public static void log(String Message, Object... objects) {
		log(defaultLevel, String.format(Message, objects));
	}

	public static void log(int level, String Message) {
		if(detailedOutput) {
			StackTraceElement[] st = Thread.currentThread().getStackTrace();
			String className = st[st.length - 1].getClassName();
			int lineNumber = st[st.length - 1].getLineNumber();
			outputMessage(level, String.format("<?%s:%d> %s \n", className,
				lineNumber, Message));
		} else {
			outputMessage(level, String.format("<??:??> %s \n",Message));
		}
	}

	public static void log(int level, String Message, Object... objects) {
		log(level, String.format(Message, objects));
	}

	private static synchronized void outputMessage(int level, String message) {
		if (outputters != null) {
			synchronized (outputters) {
				if (outputters == null)
					outputters = new ArrayList<PrintStream>();
				if (levels == null)
					levels = new HashMap<PrintStream, Integer>();
				Iterator<PrintStream> outIt = outputters.iterator();
				while (outIt.hasNext()) {
					PrintStream ps = (PrintStream) outIt.next();
					int outLevel = getLevel(ps);
					if (level <= outLevel)
						ps.print(String.format("[%d](%d)" + message, new Date()
								.getTime(), level));
				}
			}
		}
	}

	public static synchronized void addOutput(PrintStream output) {
		addOutput(defaultLevel, output);
	}

	public static synchronized void setDefaultLevel(int level) {
		defaultLevel = level;
	}

	private static int getLevel(PrintStream ps) {
		if (levels.containsKey(ps))
			return ((Integer) levels.get(ps)).intValue();
		else
			return defaultLevel;
	}

	public static synchronized void removeOutput(PrintStream output) {
		if (outputters != null) {
			synchronized (outputters) {
				if (outputters == null)
					outputters = new ArrayList<PrintStream>();
				if (levels == null)
					levels = new HashMap<PrintStream, Integer>();
				if (outputters.contains(output)) {
					output.close();
					outputters.remove(output);
					levels.remove(output);
				}
			}
		}
	}

	public static synchronized void removeAllOutputs() {
		if (outputters != null) {
			synchronized (outputters) {
				Iterator<PrintStream> outIt = outputters.iterator();
				while (outIt.hasNext()) {
					PrintStream ps = (PrintStream) outIt.next();
					ps.close();
				}
				outputters.clear();
				levels.clear();
			}
		}
	}

	public static synchronized void addOutput(int level, PrintStream output) {
		if (outputters == null)
			outputters = new ArrayList<PrintStream>();
		if (levels == null)
			levels = new HashMap<PrintStream, Integer>();
		synchronized (outputters) {
			if (!outputters.contains(output)) {
				outputters.add(output);
				levels.put(output, new Integer(level));
			}
		}
	}
	
	public static void setDetailedOutput(boolean value) {
		detailedOutput = value;
	}

}
