/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package opt.blatant;


import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;

import com.fuzzbug.solenopsis.sdk.exe.INodeDaemon;
import com.fuzzbug.solenopsis.sdk.vm.IVMState;



public class BlatAntLogic {
	
	/* Node properties */
	
	public int getMAX_OPTI_CONNECTIONS() {
		return MAX_OPTI_CONNECTIONS;
	}

	public int getMAX_CONNECTIONS() {
		return MAX_CONNECTIONS;
	}

	public int getD() {
		return D;
	}

	public int getREQUEST_MIN_INTERVAL() {
		return REQUEST_MIN_INTERVAL;
	}

	public int getALPHA_MAX_AGE() {
		return ALPHA_MAX_AGE;
	}
	
	/* ****************************************************************** */
	/* BLATANT PARAMETER HERE, BLATANT PARAMETER HERE
	/* ****************************************************************** */
	
	/* These parameters can be changed at runtime by using 
	
		(platform::call <nodeid> "blatants::setProperty" <property> <value>)
	
		For example:
	
		(platform::call id "blatants::setProperty" "MAX_CONNECTIONS" 8)
		
	*/

	private int MAX_CONNECTIONS = 8;
	private int MAX_OPTI_CONNECTIONS = 6;
	
	private int D = 5;
	
	private int REQUEST_MIN_INTERVAL = 10;
	private long CONNECTION_RETRY_INTERVAL = 10;
	private int ALPHA_MAX_AGE = 50;
	private long DISCOVERY_RESPAWN_INTERVAL = 60;
	private double DISCOVERY_RESPAWN_PROBABILITY = 0.01;

	/* ****************************************************************** */
	/* ****************************************************************** */
	
	private SAlpha alpha;
	private boolean active = false;
    private boolean isOverlayHook = false;
	private long retryConnectionIfIsolatedAt = 0;
	
	protected final Pheromone gamma;
	protected final Pheromone beta;
	
	private long timestamp = 0;
	
	protected final INodeDaemon d;
	
	private List<String> wellKnown;
	
	private final ReentrantLock lock = new ReentrantLock();
	
	public BlatAntLogic (INodeDaemon d) {
		
		
		/* ****************************************************************** */
		/* BLATANT PARAMETER HERE, BLATANT PARAMETER HERE
		/* ****************************************************************** */
		
		/* Initialize the beta and gamma pheromones:
			(String name, Double decay, Double increment, Double defvalue, Double minvalue)
		*/
		beta = new Pheromone("beta", 0.99999, 0.1, 1.0, 0.001);
		gamma = new Pheromone("gamma", 0.99999, 0.1, 1.0, 0.25);
		
		/* ****************************************************************** */
		/* ****************************************************************** */
		
		this.d = d;
		alpha = new SAlpha(this);
		wellKnown = new ArrayList<String>();
		//this.active = true;
        //this.isOverlayHook = true;
        NetworkSystem.addNode(this);
	}
	
	public final void acquire() {
		lock.lock();
	}

	public final void release() {
		lock.unlock();
	}

	public boolean connect(String target) {

		// 0. If the node is not active, reject
		if (!active) {
			return false;
		}

		// 1. If it's already a neighbor ancknowledge
		if (isNeighbor(target))
			return true;

		// 2. If we already have the maximum number of connections refuse
		if ((MAX_OPTI_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= MAX_OPTI_CONNECTIONS))
			return false;

		// 3. If its ourselves, reject
		if (target.equals(d.getIdentifier())) {
			return false;
		}

		// 4. Connect
		if (!alpha.isNeighbor(target)) {
			acquire();
			beta.fill(target);
			gamma.fill(target);
			alpha.onConnect(target);
			release();
			return true;
		}
		return false;
	}

	public boolean disconnect(String target) {
		if (alpha.isNeighbor(target)) {
			acquire();
			removeNeighbor(target);
			gamma.clear(target);
			beta.clear(target);
			alpha.onDisconnect(target);
			release();
			return true;
		}
		return true;
	}

	public int getEstimatedDistance(String target) {
		return alpha.getEstimatedDistance(target);
	}

	public void getInfosTo(Entry e) {
		alpha.getInfosTo(e);
	}

	public List<String> getNeighbors() {
		return alpha.getNeighbors();
	}

	public int getNeighborsCount() {
		return alpha.getNeighborsCount();
	}

	public boolean isNeighbor(String id) {
		return alpha.isNeighbor(id);
	}

	public boolean isActiveNeighbor(String id) {
		return alpha.isActiveNeighbor(id);
	}

	private boolean removeDeadNeighbor(String target) {
		if (alpha.isNeighbor(target)) {
			
			if (alpha.getNeighborsCount() == 1) {
                System.out.format("We will be alone... start recovery with hook!\n");
				// Initiate connection with a random hook in the network
				disconnect(target);
				initiateConnection(getWellKnownNode());
			}

			List<String> hooks = alpha.getNeighbors(target);
			if (hooks.size() <= 1) {
                System.out.format("Leaving node did not have neighbors, skipping recovery\n");
				disconnect(target);
				return true;
			}

			// Circular connection

            System.out.format("Initiating recovery\n");

			List<String> toNotify = new ArrayList<String>();
			toNotify.addAll(hooks);
			Collections.sort(toNotify, new Comparator<String>() {

				public int compare(String arg0, String arg1) {
					return arg0.toString().compareTo(arg1.toString());
				}
				
			});

			for (int t = 0; t < toNotify.size() - 1; t++) {
				// Process all but the last
				String source = toNotify.get(t);
				String dest = toNotify.get(t + 1);
                System.out.format("Sending construction recovery to %s\n", dest.toString());
				sendData(dest, AntFactory.getConstructionLinkAnt(d, source));
			}
			
			disconnect(target);

			if (getNeighborsCount() == 0) {
                if (!isOverlayHook)
                    initiateConnection(getWellKnownNode());
			}
		}
		return true;
	}

	private void initiateConnection(String hook) {
		if (hook == null) return;
		sendData(hook, AntFactory.getConstructionLinkAnt(d, d.getIdentifier()));
		retryConnectionIfIsolatedAt = getTimestamp() + CONNECTION_RETRY_INTERVAL;
	}
	
	public void startConstructionLinkAnt(String target) {
		sendData(target, AntFactory.getConstructionLinkAnt(d, d.getIdentifier()));
	}

	private void removeNeighbor(String identifier) {
		alpha.removeNeighbor(identifier);
	}

	public void tick() {
		alpha.tick();
	}
	
	public void step() {
		timestamp++;
		timestep();
	}

	private void timestep() {
		/*
		 * Check if we have been for a too long time without neighbors, if yes
		 * re-start the connection procedure we a well known node
		 */
		
		//System.out.println(getIdentifier()+"Node @ "+getTimestamp()+" isOverlayHook?"+isOverlayHook+" Neighbors?"+getNeighborsCount()+" Active?"+active+ " retryConnectionIfIsolatedAt"+retryConnectionIfIsolatedAt);
		
		if (!isOverlayHook && (getTimestamp() >= retryConnectionIfIsolatedAt)) {
			if ((this.getNeighborsCount() == 0) || !active) {
				initiateConnection(getWellKnownNode());
			}
		}
		
		/*
		 * Respawn discovery ant with some probability
		 */
		if (getTimestamp() % DISCOVERY_RESPAWN_INTERVAL == 0) {
			if (Math.random() < DISCOVERY_RESPAWN_PROBABILITY) {
				if (getNeighborsCount() != 0) {
					sendData(this.getIdentifier(), AntFactory.getDiscoveryAnt(d));
				}
			}
		}
		
		/*
		 * If the gamma pheromone (out-pheromone) is low, not enough traffic did
		 * go on the corresponding link. So we need to send a ping.
		 */
		List<String> g = gamma.decay();
		for (String n : g) {
			if (isNeighbor(n)) {
				sendData(n, AntFactory.getPingAnt(d, d.getIdentifier()));
			}
		}

		/*
		 * If the beta pheromone (in-pheromone) is low, not enough traffic
		 * arrived from a link. We initiate a recovery procedure.
		 */
		List<String> b = beta.decay();
		for (String n : b) {
			if (!alpha.isActiveNeighbor(n)) { // Node never communicate with
												// us
                System.out.format("Lost neighbor %s was not active\n", n.toString());
				disconnect(n);
			} else if (alpha.isNeighbor(n) && alpha.isActiveNeighbor(n)) { // Node															// dead
                System.out.format("Lost active neighbor %s: {%s}!\n", n.toString(), alpha.getNeighbors(n));
				removeDeadNeighbor(n);
			} else {
                System.out.format("Unknown node %s disconnected. Take no action.\n", n.toString());
            }
		}
	}

	private String getWellKnownNode() {
		if (wellKnown.isEmpty()) return null;
		return wellKnown.get((int) (wellKnown.size() * Math.random()));
	}
	
	public void addWellKnown(String address) {
		wellKnown.add(address);
	}
	
	public boolean removeWellKnown(String address) {
		return wellKnown.remove(address);
	}

	public void updateEntry(Entry e, int d) {
		acquire();
		alpha.updateEntry(e, d);
		release();
	}

	public void activateNeighbor(String neighbor) {
		acquire();
		alpha.activateNeighbor(neighbor);
		release();
	}

	public void updateInformation(List<Entry> visited) {
		acquire();
		alpha.updateInformation(visited);
		release();
	}

	public void updateNeighbors(String neighbor, Long timestamp, List<String> ns) {
		acquire();
		alpha.updateNeighbors(neighbor, timestamp, ns);
		release();
	}

	public boolean userconnect(String target, int ttl) {

		// 0. If the node is not active, reject
		if (!active) {
			return false;
		}

		// 1. Check if we are not already neighbors
		if (isNeighbor(target)) {
			return true;
		}

		// 2. Check if we are connecting to ourselves
		if (target.equals(d.getIdentifier())) {
			return false;
		}

		// 3. Check if we have still some free connections
		if ((MAX_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= MAX_CONNECTIONS)) {
			return false;
		}

		// 4. If the ant ttl is > 0, check if we are full of optimization
		// connectios
		if ((ttl > 0)
				&& (MAX_OPTI_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= MAX_OPTI_CONNECTIONS)) {
			return false;
		}

		// 5. If the node is not active, leave a free connection for the
		// activating connection
		if ((!active)
				&& (MAX_OPTI_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= MAX_OPTI_CONNECTIONS - 1)) {
			return false;
		}

		// 6. Accept the connection
		acquire();
		beta.fill(target); // if beta is not filled here, if the
										// ant gets lost in its
		// way to the target (the source node requesting connection)
		// then we have a neighbor that is not tracked by a beta
		// pheromone, meaning that we lose one connection and we
		// are not able to notice it.
		gamma.fill(target);
		alpha.onConnect(target);
		release();
		return true;
	}

	/**
	 * Second step user connection (when OUR construction-link ant comes back)
	 */
	public boolean userconnect(String target) {

		// 1. Check if we are not already neighbors
		if (isNeighbor(target)) {
			return true;
		}

		// 2. Check if the request does not come from ourselves
		if (target.equals(d.getIdentifier())) {
			return false;
		}

		// 3. Check if we have enough free connections
		if ((MAX_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= MAX_CONNECTIONS)) {
			return false;
		}

		// 4. Accept the connection
		acquire();
		beta.fill(target);
		gamma.fill(target);
		alpha.onConnect(target);
		active = true;
		release();
		
		activateNeighbor(target); // Since it answered (we are at step 2) it
									// can be considered as active
		
		sendData(target, AntFactory.getPingAnt(d, d.getIdentifier()));

		return true;
	}

	public int getNeighborNeighborsSize(String n) {
		return alpha.getNeighbors(n).size();
	}

	public void cleanLeave() { // Create circular connection between neighbors
		if (getNeighborsCount() <= 1) // Only one neighbor, or none, do
										// nothing
			return;
		List<String> toNotify = new ArrayList<String>();
		toNotify.addAll(getNeighbors());
		Collections.sort(toNotify, new Comparator<String>() {

			public int compare(String arg0, String arg1) {
				return arg0.toString().compareTo(arg1.toString());
			}
			
		});
		if (toNotify.size() > 2) {
			String last = toNotify.get(toNotify.size() - 1);
			toNotify.add(0, last);
		}

		for (String ns : getNeighbors()) {
			sendData(ns, AntFactory.getUnLinkAnt(d, d.getIdentifier()));
		}

		for (int t = 0; t < toNotify.size() - 1; t++) {
			System.out.println("Clean leave");
			// Process all but the last
			String source = toNotify.get(t);
			String dest = toNotify.get(t + 1);
			sendData(dest, AntFactory.getConstructionLinkAnt(d, source));
		}
	}
	
	public final Pheromone getGamma() {
		return gamma;
	}

	public final Pheromone getBeta() {
		return beta;
	}

	public final String getIdentifier() {
		synchronized(d.getIdentifier()) {
			return d.getIdentifier();
		}
	}

	public long getTimestamp() {
		return timestamp;
	}

	public void sendData(String recipient, IVMState s) {
			gamma.fill(recipient);
			d.sendState(recipient, s);
	}

	public void setActive(boolean b) {
		active = b;
	}
	
	public void setIsWellKnown(boolean b) {
		isOverlayHook = b;
	}
}
