/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package opt.blatant;


import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import com.fuzzbug.solenopsis.sdk.exe.INodeDaemon;
import com.fuzzbug.solenopsis.sdk.plugins.BasicPlugin;
import com.fuzzbug.solenopsis.sdk.scheduling.IAlarmable;
import com.fuzzbug.solenopsis.sdk.vm.IVM;


public class BlatAntProxy extends BasicPlugin implements IAlarmable {
	
	private BlatAntLogic bl;
	
	private enum CALLBACK { TIMESTEP };
	
	public BlatAntProxy(INodeDaemon d) {
		super("blatants", "Amos Brocco <amos.brocco@unifr.ch", "BlatAntS Proxy", "1.0", false);
		registerMethods();
		bl = new BlatAntLogic(d);
		d.getNodeAlarmService().schedule(this, CALLBACK.TIMESTEP, 1000);
	}

	@Override
	public void callback(UUID taskId, Object data, long timestamp) {
		if (data == CALLBACK.TIMESTEP) {
			bl.step();
		}
	}
	
	/*
	 * Methods
	 */
	
	public Object apply_getIdentifier(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		return bl.getIdentifier();
	}
	
	public Object apply_getNeighbors(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		return bl.getNeighbors();
	}
	
	public Object apply_getEstimatedDistance(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		String n = (String) args.get(0);
		return bl.getEstimatedDistance(n);
	}
	
	public Object apply_activateNeighbor(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		String neighbor = (String) args.get(0);
		bl.activateNeighbor(neighbor);
		return 1;
	}
	
	public Object apply_setActive(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		boolean value = Boolean.FALSE.equals(args.get(0));
		bl.setActive(!value);
		return value;
	}
	
	public Object apply_setIsWellKnown(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		boolean value = Boolean.FALSE.equals(args.get(0));
		bl.setIsWellKnown(!value);
		return value;
	}
	
	public Object apply_getGamma(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		String trail = (String) args.get(0);
		return bl.getGamma().senseTrail(trail);
	}
	
	public Object apply_getBeta(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		String trail = (String) args.get(0);
		return bl.getBeta().senseTrail(trail);
	}
	
	public Object apply_fillGamma(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		String trail = (String) args.get(0);
		bl.getGamma().fill(trail);
		return 1;
	}
	
	public Object apply_fillBeta(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		String trail = (String) args.get(0);
		bl.getBeta().fill(trail);
		return 1;
	}
	
	@SuppressWarnings("unchecked")
	public Object apply_inform(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		List<Map<String,Object>> list = (List<Map<String,Object>>) args.get(0);
		List<Entry> elist = new ArrayList<Entry>();
		for (Map<String,Object> m : list) {
			Entry e = new Entry((String) m.get("id"));
			e.updateFrom(m);
			elist.add(e);
		}
		bl.updateInformation(elist);
		bl.tick();
		return false;
	}
	
	@SuppressWarnings("unchecked")
	public Object apply_updateEntry(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 2) return 0;
		Map<String,Object> m = (Map<String,Object>) args.get(0);
		Entry e = new Entry((String) m.get("id"));
		Integer d = (Integer) args.get(0);
		e.updateFrom(m);
		bl.updateEntry(e, d);
		bl.tick();
		return false;
	}
	
	public Object apply_tick(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		bl.tick();
		return 1;
	}
	
	public Object apply_connect(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		String target = (String) args.get(0);
		return bl.connect(target) ? 1 : false;
	}
	
	public Object apply_userConnect(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() < 1) return 0;
		String target = (String) args.get(0);
		if (args.size() > 1) {
			int ttl = (Integer) args.get(1);
			return bl.userconnect(target, ttl) ? 1 : false;
		} else
			return bl.userconnect(target) ? 1 : false;
	}
	
	public Object apply_constructionConnect(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() < 1) return 0;
		String target = (String) args.get(0);
		bl.startConstructionLinkAnt(target);
		return false;
	}

	public Object apply_disconnect(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		String target = (String) args.get(0);
		return bl.disconnect(target) ? 1 : false;
	}
	
	public Object apply_getInformation(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		Entry e = new Entry(bl.getIdentifier());
		bl.getInfosTo(e);
		Map<String,Object> bv = e.toBaseValue();
		bv.remove("neighbors");
		return bv;
	}
	
	@SuppressWarnings("unchecked")
	public Object apply_updateNeighbors(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 3) return 0;
		String target = (String) args.get(0);
		Number ts = (Number) args.get(1);
		List<String> ns = (List<String>) args.get(2);
		bl.updateNeighbors(target, ts.longValue(), ns);
		return 1;
	}
	public Object apply_getStats(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		int cpus = Runtime.getRuntime().availableProcessors();
		ParallelCompute pcomputer = new ParallelCompute(cpus);
		pcomputer.computeBFS();
		int diameter = pcomputer.getDiameter();
		int radius = pcomputer.getRadius();
		float avpl = pcomputer.getAveragePathLength();
		int largest = pcomputer.largest();
		int smallest = pcomputer.smallest();
		int size = NetworkSystem.getNodeCount();
		int isolated = NetworkSystem.getIsolatedCount();
		int edges = NetworkSystem.getEdgeCount();
		Map<String,Object> stats = new HashMap<String,Object>();
		stats.put("diameter", diameter);
		stats.put("radius", radius);
		stats.put("avpl", avpl);
		stats.put("largest", largest);
		stats.put("smallest", smallest);
		stats.put("size", size);
		stats.put("isolated", isolated);
		stats.put("edges", edges);
		return stats;
	}
	
	public BlatAntLogic getLogic() {
		return bl;
	}
	
	public Object apply_addWellKnown(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		bl.acquire();
		String hook = (String) args.get(0);
		bl.addWellKnown(hook);
		bl.release();
		return hook;
	}
	
	public Object apply_removeWellKnown(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		bl.acquire();
		String hook = (String) args.get(0);
		Object ret = bl.removeWellKnown(hook) ? 1 : false;
		bl.release();
		return ret;
	}
	
	public Object apply_cleanLeave(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		bl.cleanLeave();
		return 1;
	}
	
	public Object apply_getTimestamp(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		return bl.getTimestamp();
	}
	
	public Object apply_setProperty(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 2) return 0;
		String property = (String) args.get(0);
		Number value = (Number) args.get(1);
		try {
			Field f = BlatAntLogic.class.getDeclaredField(property);
			f.setAccessible(true);
			f.set(bl, value);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
			return false;
		}
		return value;
	}
	
	public Object apply_getProperty(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return 0;
		String property = (String) args.get(0);
		Object value = false;
		try {
			Field f = BlatAntLogic.class.getDeclaredField(property);
			f.setAccessible(true);
			value = f.get(bl);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return value;
	}
	
}
