/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package opt.blatant;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.fuzzbug.solenopsis.sdk.lib.IBaseSerializable;


public class Entry implements IBaseSerializable {

	private static final long serialVersionUID = 1949656046021960318L;
	private String identifier = null;
	private Set<String> neighbors;
	private int distance;
	private int computed;
	private long entryAge;
	private long remoteTime;
	private boolean sticky = false;
	private boolean free = false;
	private long nextRequestMinTimestamp = 0;

	public Entry(String id) {
		if (id != null)
			identifier = id;  
		neighbors = new HashSet<String>();
		distance = -1;
		computed = -1;
		entryAge = 0;
		remoteTime = 0;
		nextRequestMinTimestamp = 0;
	}

	public void setNextRequestMinTimestamp(long timestamp) {
		nextRequestMinTimestamp = timestamp;
	}

	public long getNextRequestMinTimestamp() {
		return nextRequestMinTimestamp;
	}

	public synchronized void addNeighbor(String id) {
		neighbors.add(id);
	}

	public synchronized void clear() {
		neighbors.clear();
	}

	public synchronized void clearNeighbors() {
		neighbors.clear();
	}

	public void free() {
		free = true;
	}

	public int getComputedDistance() {
		return computed;
	}

	public int getDistance() {
		return distance;
	}

	public long getEntryAge() {
		return entryAge;
	}

	public String getId() {
		return identifier;
	}

	public synchronized List<String> getNeighbors() {
		List<String> ns = new ArrayList<String>(neighbors.size());
		getNeighbors(ns);
		return ns;
	}

	public synchronized void getNeighbors(List<String> dest) {
		for (String n : neighbors)
			dest.add(n);
	}

	public synchronized void getNeighbors(Set<String> dest) {
		for (String n : neighbors)
			dest.add(n);
	}

	public synchronized int getNeighborsCount() {
		return neighbors.size();
	}

	public synchronized int getNeighborsHash() {
		return neighbors.hashCode();
	}

	public long getRemoteTime() {
		return remoteTime;
	}

	public boolean isFree() {
		return free;
	}

	public synchronized boolean isNeighbor(String id) {
		return neighbors.contains(id);
	}

	public boolean isSticky() {
		return sticky;
	}

	public synchronized void removeNeighbor(String id) {
		neighbors.remove(id);
	}

	public void setComputedDistance(int d) {
		computed = d;
	}

	public void setDistance(int d) {
		distance = d;
	}

	public void setEntryAge(long a) {
		entryAge = a;
	}

	public void setId(String id) {
		identifier = id;
	}

	public void setRemoteTime(long a) {
		remoteTime = a;
	}

	public void stick() {
		sticky = true;
	}

	public long tick() {
		remoteTime++;
		return remoteTime;
	}

	public String toString() {
		return String.format(" (id=%s,ns=%s,d=%d,cd=%d,age=%d,rt=%d) ",
				identifier.toString(), neighbors.toString(), distance, computed, entryAge,
				remoteTime);
	}

	public void unstick() {
		sticky = false;
	}

	public synchronized boolean updateFrom(Entry e) {
		if (e == null)
			return false;

		identifier = e.getId();
		remoteTime = new Long(e.remoteTime);

		if (neighbors.hashCode() != e.getNeighborsHash()) {
			neighbors.clear();
			e.getNeighbors(neighbors);
			return true;
		}

		return false;
	}
	
	@SuppressWarnings("unchecked")
	public synchronized void updateFrom(Map<String, Object> map) {
		if (map.containsKey("id"))
			identifier = (String) map.get("id");
		if (map.containsKey("time"))
			remoteTime = (Long) map.get("time");
		if (map.containsKey("neighbors")) {
			neighbors.clear();
			neighbors.addAll((List<String>) map.get("neighbors"));
		}
	}
	
	public synchronized Map<String, Object> toBaseValue() {
		Map<String,Object> map = new HashMap<String,Object>();
		map.put("id", identifier);
		map.put("time", remoteTime);
		List<String> ns = new ArrayList<String>();
		ns.addAll(neighbors);
		map.put("neighbors", ns);
		return map;
	}

}
