/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package opt.blatant;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.fuzzbug.solenopsis.sdk.vm.IVMState;



public class SAlpha {

	private String identifier;
	private BlatAntLogic owner;
	private Table table;
	private boolean changed = false;
	private Entry localInfo;
	

	public SAlpha(BlatAntLogic o) {
		table = new Table(40);
		owner = o;
		identifier = owner.getIdentifier();
		localInfo = new Entry(identifier);
	}
	
	private void neighborsChanged() {
		IVMState p = AntFactory.getUpdateNeighborsAnt(owner.d, identifier, owner.getTimestamp(), getNeighbors());
		for (String n : getNeighbors()) {
			owner.sendData(n, p);
		}
	}

	public void addNeighbor(String id) {
		if (!localInfo.isNeighbor(id)) {
			localInfo.addNeighbor(id);
			table.stick(table.getIndexOrNew(id));
			neighborsChanged();
			onPeerConnected(id);
		}
	}

	public void clear() {
		table.clear();
		localInfo.clear();
	}

	/**
	 * Evaluates a connection to another node
	 */
	private void evaluateConnection() {
		// /////////////////////////////////////////////////////////
		// PRECONDITION, WE CAN ADD NEW CONNECTIONS
		// /////////////////////////////////////////////////////////
		if ((owner.getMAX_OPTI_CONNECTIONS() > 0)
				&& (getNeighborsCount() >= owner.getMAX_OPTI_CONNECTIONS()))
			return;

		// /////////////////////////////////////////////////////////
		// EVALUATE THE DISTANCES
		// /////////////////////////////////////////////////////////

		int max_distance = 0;
		String max_distance_node = null;

		for (Entry e : table) {
			if (e == null) // Non valid
				continue;
			if (e.isFree()) // Non valid
				continue;
			if (e.isSticky()) // Neighbor
				continue;
			if ((owner.getTimestamp() - e.getEntryAge()) > owner.getALPHA_MAX_AGE()) {// Too old
				if (!e.isSticky())
					e.free();
				continue;
			}

			int estimated = getEstimatedDistance(e.getId());

			if (estimated < 0)
				continue;
			if (estimated < (2 * owner.getD() - 1)) // Distance is not
														// sufficient
				continue;
			if (e.getNeighborsCount() >= owner.getMAX_OPTI_CONNECTIONS()) // Don't
																			// test
																			// if
																			// target
																			// is
																			// full
				continue;
			if (owner.getTimestamp() < e.getNextRequestMinTimestamp()) // If
																		// it's
																		// not
																		// yet
																		// time
																		// to
																		// repeat
																		// the
																		// request
																		// don't
																		// do it
				continue;

			if (estimated > max_distance) {
				max_distance = estimated;
				max_distance_node = e.getId();
			}
		}

		if (max_distance >= (2 * owner.getD() - 1)) { // >= 2D - 1
			table.get(max_distance_node).setNextRequestMinTimestamp(
					owner.getTimestamp() + max_distance
							* owner.getREQUEST_MIN_INTERVAL());
			//System.out.format("Sending Optimization Ant to '%s'\n", max_distance_node.toString());
			owner.sendData(max_distance_node, AntFactory.getOptimizationLinkAnt(owner.d, identifier));
		}
	}
	
	private boolean isSuperior(String a, String b) {
		return a.hashCode() > b.hashCode();
	}

	/**
	 * Evaluates a disconnection from a neighbor
	 */
	private void evaluateDisconnection(List<Entry> info) {

		int d = 0;
		String last = null;

		for (Entry e : info) {

			// If we hit ourselves, reset the data
			if (e.getId().equals(identifier)) {
				last = null;
				d = 0;
				continue;
			}

			// If we encounter a node with a superior id, reset
			if (isSuperior(e.getId(), identifier)) {
				last = null;
				d = 0;
				continue;
			}

			// If we encounter an active neighbor
			if (isActiveNeighbor(e.getId())) {
				// ... and there was already a neighbor before
				if (last != null) {
					if (last.equals(e.getId())) {
						d = 0;
					} else if (d <= (2*owner.getD() - 3)) { // ... <= 2D-3, ie < 2D-2 and
																	// the
																	// distance
																	// is less
																	// than the
																	// limit,
																	// disconnect
						int lastNScount = getNeighbors(last).size();
						int thisNScount = getNeighbors(e.getId()).size();
						// Disconnect the neighbor with the highest degree
						// ... so that we help overloaded nodes
						if (lastNScount > thisNScount) {
							this.owner.disconnect(last); // Free the
															// connection as
															// soon as possible
							owner.sendData(last, AntFactory.getUnLinkAnt(owner.d, identifier));
							last = null;
							d = 0;
							continue; // FIXME: why this was a break?
						} else {
							this.owner.disconnect(e.getId()); // Free the
																// connection as
																// soon as
																// possible
							owner.sendData(e.getId(), AntFactory.getUnLinkAnt(owner.d, identifier));
							last = null;
							d = 0;
							continue; // FIXME: why this was a break?
						}
					} else {
						// ... reset all parameters to this new neighbor
						d = 0;
						last = e.getId();
					}
				} else {
					// we didn't have a last, set it
					d = 0;
					last = e.getId();
				}
			}
			d++;
		}
	}

	public int getEstimatedDistance(String target) {
		if (target.equals(identifier))
			return 0;

		if (isNeighbor(target))
			return 1;
		
		for(String t : getNeighbors()) {
			Entry en = table.get(t);
			if (en != null) {
				if (en.isNeighbor(target))
					return 2;
			}
		}
		
		Entry e = table.get(target);
		if (e == null)
			return -1;

		return e.getDistance();
	}

	public void getInfosTo(Entry e) {
		e.updateFrom(localInfo);
		e.setRemoteTime(owner.getTimestamp());
	}

	public List<String> getNeighbors() {
		return localInfo.getNeighbors();
	}

	public List<String> getNeighbors(String target) {
		Entry e = table.get(target);
		if (e != null) {
			return e.getNeighbors();
		} else {
			return new ArrayList<String>();
		}
	}

	public Integer getNeighborsCount() {
		return localInfo.getNeighborsCount();
	}

	public void getNeighborsTo(Set<String> ns) {
		localInfo.getNeighbors(ns);
	}

	public boolean isActiveNeighbor(String id) {
		return localInfo.isNeighbor(id) && table.contains(id)
				&& table.get(id).getNeighbors().contains(identifier);
	}

	public boolean isNeighbor(String id) {
		return localInfo.isNeighbor(id);
	}

	public void onConnect(String target) {
		if (!localInfo.isNeighbor(target)) {
			addNeighbor(target);
			int i = table.getIndexOrNew(target);
			Entry e = table.get(i);
			e.setDistance(1);
		}
	}

	public void onDisconnect(String target) {
		if (localInfo.isNeighbor(target)) {
			Entry e = table.get(target);
			if (e != null) {
				e.unstick();
				e.removeNeighbor(identifier);
				e.setDistance(owner.getD());
			}
		}
	}

	public void removeNeighbor(String id) {
		if (localInfo.isNeighbor(id)) {
			int i = table.getIndex(id);
			table.unstick(i);
			localInfo.removeNeighbor(id);
			neighborsChanged();
			onPeerDisconnected(id);
		}
	}
	
	private void onPeerConnected(String n) {
	}
	
	private void onPeerDisconnected(String n) {
	}

	public void tick() {
		localInfo.tick();
		if (changed) {
				evaluateConnection();
				changed = false;
		}
	}

	public void updateEntry(Entry e, int d) {
		Entry entry = table.getOrNew(e.getId());
		long ts = entry.getRemoteTime();
		if (e.getRemoteTime() > ts) {
			if (entry.updateFrom(e))
				changed = true;
			entry.setEntryAge(localInfo.getRemoteTime());
			if ((d == 1) && !isNeighbor(e.getId()))
				return;
			entry.setDistance(d);
		}
	}

	public void updateInformation(List<Entry> info) {
		int distanceFromNeighbor = 2; // We start from the first node after the neighbor
		Map<String, Integer> distances = new HashMap<String, Integer>();
		Map<String, Long> timestamps = new HashMap<String, Long>();

		for (int i = info.size() - 2; i >= 0; i--) {
			Entry item = info.get(i);
			String id = item.getId();
			long ts = item.getRemoteTime();
			if (isNeighbor(id)) {
				distanceFromNeighbor = 2;
			} else {
				distances.put(id, distanceFromNeighbor);
				if (!timestamps.containsKey(id))
					timestamps.put(id, ts);
				distanceFromNeighbor++;
			}
		}

		distanceFromNeighbor = info.size(); // We start at the end of the vector 
												   // (we don't have a neighbor to evaluate, so estimate
												   // to the length of the vector
		
		for (int i = 0; i < info.size() - 2; i++) {
			Entry item = info.get(i);
			String id = (String) item.getId();
			if (isNeighbor(id)) {
				distanceFromNeighbor = 2;
			} else {
				int value = Math.min(distanceFromNeighbor, distances.get(id));
				distances.put(id, value);
				distanceFromNeighbor++;
			}
		}

		for (String k : distances.keySet()) {
			Entry entry = table.getOrNew(k);
			long ts = entry.getRemoteTime();
			if (timestamps.get(k) > ts) {
				entry.setRemoteTime(timestamps.get(k));
				entry.setEntryAge(owner.getTimestamp());
				entry.setDistance(distances.get(k));
			} else if (timesAreClose(timestamps.get(k), ts)) { // If the
																// timestamp is
																// "similar"
																// take the
																// smallest
																// value
				entry.setRemoteTime(timestamps.get(k));
				entry.setEntryAge(owner.getTimestamp());
				int dist = entry.getDistance();
				if (dist < 0) {
					entry.setDistance(distances.get(k));
				} else {
					entry.setDistance(Math.min(dist, distances.get(k)));
				}
			}
			changed = true;
		}

		evaluateDisconnection(info);
	}

	private boolean timesAreClose(Long a, long b) {
		if (Math.abs(b - a) < 5) {
			return true;
		} else
			return false;
	}

	public void updateNeighbors(String neighbor, Long timestamp, List<String> ns) {
		Entry e = table.getOrNew(neighbor); // Avoid missing our own connection
		if (e != null) {
            long ts = e.getRemoteTime();
            if (timestamp >= ts) {
                //System.out.format("Updating neighbors of %s to %s\n", neighbor.toString(), ns.toString());
                e.clearNeighbors();
                for (String n : ns)
                    e.addNeighbor(n);
            } else {
            	//System.out.format("Updating neighbors of %s skipped\n", neighbor.toString());
            }
            e.setRemoteTime(Math.max(ts,timestamp));
            e.setEntryAge(owner.getTimestamp());
		}
	}

	public int getTimestamp() {
		return (int) localInfo.getRemoteTime();
	}

	public void activateNeighbor(String neighbor) {
		if (!isNeighbor(neighbor))
			return;
		Entry n = table.get(neighbor);
		if ((n != null) && !n.isNeighbor(identifier))
			n.addNeighbor(identifier);
	}

}
