/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package scheduling;

import interfaces.scheduling.IAlarmServiceProvider;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.UUID;

import com.fuzzbug.solenopsis.sdk.scheduling.IAlarmable;


public class AlarmService implements IAlarmServiceProvider {
	
	private LinkedList<AlarmTasklet> schedule;
	
	private Object mutex;
	
	private long logicTime = 0l;
	
	private boolean useRealTime = false;
	
	public AlarmService(boolean useRealtime) {
		schedule = new LinkedList<AlarmTasklet>();
		mutex = new Object();
		this.useRealTime = useRealtime;
	}

	public UUID[] getSchedule(IAlarmable client) {
		synchronized(mutex) {
			LinkedList<UUID> ids = new LinkedList<UUID>();
			Iterator<AlarmTasklet> iter = schedule.iterator();
			while(iter.hasNext()) {
				AlarmTasklet task = iter.next();
				if(task.obj == client)
					ids.add(task.identifier);
			}
			return (UUID[]) ids.toArray();
		}
	}

	public boolean hasSchedule(IAlarmable client) {
		synchronized(mutex) {
			Iterator<AlarmTasklet> iter = schedule.iterator();
			while(iter.hasNext()) {
				AlarmTasklet task = iter.next();
				if(task.obj == client)
					return true;
			}
			return false;
		}
	}

	public UUID schedule(IAlarmable client, Object data, long interval, boolean async) {
		synchronized(mutex) {
			AlarmTasklet task = new AlarmTasklet();
			task.interval = interval;
			task.async = async;
			task.obj = client;
			task.data = data;
			task.reschedule = true;
			task.timestamp = getTime() + interval;
			task.identifier = UUID.randomUUID();
			int position = getSchedulingPosition(task.timestamp);
			if(position >= schedule.size())
				schedule.addLast(task);
			else
				schedule.add(position, task);
			return task.identifier;
		}
	}

	public UUID scheduleAt(IAlarmable client, Object data, long timestamp, boolean async) {
		synchronized(mutex) {
			AlarmTasklet task = new AlarmTasklet();
			task.interval = 0;
			task.async = async;
			task.obj = client;
			task.data = data;
			task.reschedule = false;
			task.timestamp = timestamp;
			task.identifier = UUID.randomUUID();
			int position = getSchedulingPosition(task.timestamp);
			if(position >= schedule.size())
				schedule.addLast(task);
			else
				schedule.add(position, task);
			return task.identifier;
		}
	}
	
	public UUID scheduleAfter(IAlarmable client, Object data, long interval, boolean async) {
		synchronized(mutex) {
			long timestamp = getTime() + interval;
			AlarmTasklet task = new AlarmTasklet();
			task.interval = 0;
			task.async = async;
			task.obj = client;
			task.data = data;
			task.reschedule = false;
			task.timestamp = timestamp;
			task.identifier = UUID.randomUUID();
			int position = getSchedulingPosition(task.timestamp);
			if(position >= schedule.size())
				schedule.addLast(task);
			else
				schedule.add(position, task);
			return task.identifier;
		}
	}

	public void unschedule(UUID taskId) {
		synchronized(mutex) {
			LinkedList<AlarmTasklet> newSchedule = new LinkedList<AlarmTasklet>();
			Iterator<AlarmTasklet> iter = schedule.iterator();
			while(iter.hasNext()) {
				AlarmTasklet task = iter.next();
				if(!task.identifier.equals(taskId))
					newSchedule.add(task);
			}
			schedule = newSchedule;
		}
	}

	public void unscheduleAll(IAlarmable client) {
		synchronized(mutex) {
			LinkedList<AlarmTasklet> newSchedule = new LinkedList<AlarmTasklet>();
			Iterator<AlarmTasklet> iter = schedule.iterator();
			while(iter.hasNext()) {
				AlarmTasklet task = iter.next();
				if(task.obj != client)
					newSchedule.add(task);
			}
			schedule = newSchedule;
		}
	}
	
	private void reschedule(UUID taskId) {
			int position = getTaskletIndex(taskId);
			AlarmTasklet task = (AlarmTasklet) schedule.remove(position);
			task.timestamp = new Long(getTime() + task.interval);
			position = getSchedulingPosition(task.timestamp);
			if(position >= schedule.size())
				schedule.addLast(task);
			else
				schedule.add(position, task);
	}
	
	private int getSchedulingPosition(long timestamp) {
		int listSize = schedule.size();
		if(timestamp <= getFirstSchedulingTimestamp())
			return 0;
		if(timestamp >= getLastSchedulingTimestamp())
			return listSize;
		for(int i=0; i<listSize; i++) {
			AlarmTasklet task = (AlarmTasklet) schedule.get(i);
			if(task.timestamp > timestamp)
				return i - 1;
		}
		return listSize;
	}
	
	private long getFirstSchedulingTimestamp() {
		if(schedule.size() < 1)
			return -1;
		AlarmTasklet task = (AlarmTasklet) schedule.getFirst();
		return task.timestamp;
	}
	
	private long getLastSchedulingTimestamp() {
		if(schedule.size() < 1)
			return -1;
		AlarmTasklet task = (AlarmTasklet) schedule.getLast();
		return task.timestamp;
	}
	
	private int getTaskletIndex(UUID taskId) {
		int listSize = schedule.size();
		for(int i=0; i<listSize; i++) {
			AlarmTasklet task = (AlarmTasklet) schedule.get(i);
			if(task.identifier.equals(taskId))
				return i;
		}
		return -1;
	}

	public void processAlarms() {
		updateTime();
		LinkedList<AlarmTasklet> toBeRescheduled = new LinkedList<AlarmTasklet>();
		LinkedList<AlarmTasklet> toBeExecuted = new LinkedList<AlarmTasklet>();
		LinkedList<AlarmTasklet> toBeTrashed = new LinkedList<AlarmTasklet>();
		synchronized(mutex) {
			if(getTime() < getFirstSchedulingTimestamp())
				return;
			for (AlarmTasklet task : schedule) {
				if (task.timestamp <= getTime()) {
					toBeExecuted.add(task);
					if(task.reschedule)
						toBeRescheduled.add(task);
					else {
						toBeTrashed.add(task);
					}
				} else {
					break;
				}
			}
			schedule.removeAll(toBeTrashed);
			for (AlarmTasklet tasklet : toBeRescheduled) {
				reschedule(tasklet.identifier);
			}				
		}
		/* Execute outside mutex to avoid deadlocks with sync calls */
		Iterator<AlarmTasklet> executionIter = toBeExecuted.iterator();
		while(executionIter.hasNext()) {
			AlarmTasklet task = executionIter.next();
			if(task.async) {
				task.setCurrentTimestamp(getTime());
				new Thread(task).start();
			} else {
				task.obj.callback(task.identifier, task.data, getTime());
			}
		}
	}

	private void updateTime() {
		if (!useRealTime)
			logicTime++;
	}

	public UUID scheduleIn(IAlarmable client, Object data, long interval, long firsttime, boolean async) {
		synchronized(mutex) {
			AlarmTasklet task = new AlarmTasklet();
			task.interval = interval;
			task.async = async;
			task.obj = client;
			task.data = data;
			task.reschedule = true;
			task.timestamp = getTime() + firsttime;
			task.identifier = UUID.randomUUID();
			int position = getSchedulingPosition(task.timestamp);
			if(position >= schedule.size())
				schedule.addLast(task);
			else
				schedule.add(position, task);
			return task.identifier;
		}
	}
	
	public long getTime() {
		if (useRealTime) {
			return System.currentTimeMillis();
		} else
			return logicTime;
	}

}
