/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package serialization;

import interfaces.vm.util.IDeSerializer;
import interfaces.vm.valuetypes.ILambdaValue;
import interfaces.vm.valuetypes.IReferenceValue;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import com.fuzzbug.solenopsis.sdk.lib.IConstants;
import com.fuzzbug.solenopsis.sdk.lib.IConstants.OPCODES;

import core.vm.lib.VMCode;
import core.vm.valuetypes.LambdaValue;
import core.vm.valuetypes.ReferenceValue;
import dot.codegenerator.Instruction;

public class DeSerializer implements IDeSerializer {

	private InputStreamReader streamReader;

	private Object dataTree;

	private char ch;

	private boolean skipNextRead = false;
	
	private List<Object> directory = new ArrayList<Object>();;

	public DeSerializer(InputStream serialized) throws IOException {
			streamReader = new InputStreamReader(serialized);
			dataTree = readValue();
			serialized.close();
			streamReader.close();
	}
	
	private char read() throws IOException {
		if(!skipNextRead) {
			ch = (char) streamReader.read();
		} else
			skipNextRead = false;
		return ch;
	}
	
	private void unread() {
		skipNextRead = true;
	}

	private Object readValue() throws IOException {
		read();
		if (ch == 'P') {
			return readPointer();
		} else {
			Object ret;
			if (Character.isDigit(ch)) {
				return  readString();
			} else if (ch == 'c') {
				return  readOpCode();
			} else if (ch == 'i') {
				return  readInteger();
			} else if (ch == 'I') {
				return  readLong();
			} else if (ch == 'f') {
				return readFloat();
			} else if (ch == 'F') {
				return  readDouble();
			} else if (ch == 'd') {
				ret =  readDictionary();
			} else if (ch == 's') {
				ret =  readStack(); 
			} else if (ch == 'l') {
				ret =  readList();
			} else if (ch == 'o') {
				return  readBoolean();
			} else if (ch == 'n') {
				return  readNil();
			} else if (ch == 'x') {
				ret =  readLambda();
			} else if (ch == 'r') {
				return readReference();
			} else if (ch == 'X') {
				return  readInstruction();
			} else if (ch == 'N') {
				read();
				return null;
			} else {
				return null;
			}
			directory.add(ret);
			return ret;
		}
	}

	@SuppressWarnings("unchecked")
	private Object readInstruction() throws IOException {
		List<Object> data = (List<Object>) readValue();
		read();
		return new Instruction(data);
	}

	private Object readOpCode() throws IOException {
		StringBuilder sb = new StringBuilder();
		read();
		while (ch != 'e') {
			sb.append(ch);
			read();
		}
		Integer op = Integer.parseInt(sb.toString());
		return (OPCODES) IConstants.OPCODES.values()[op];
	}

	private List<Object> readList() throws IOException {
		ArrayList<Object> list = new ArrayList<Object>();
		read();
		if (ch == 'e')
			return list;
		else
			unread();
		while (true) {
			Object item = readValue();
			if (item == null)
				break;
			list.add(item);
			read();
			unread();
		}
		return list;
	}
	
	private Stack<Object> readStack() throws IOException {
		Stack<Object> stack = new Stack<Object>();
		read();
		if (ch == 'e')
			return stack;
		else
			unread();
		while (true) {
			Object item = readValue();
			if (item == null)
				break;
			stack.add(item);
			read();
			unread();
		}
		return stack;
	}

	private Integer readInteger() throws IOException {
		StringBuilder sb = new StringBuilder();
		read();
		while (ch != 'e') {
			sb.append(ch);
			read();
		}
		return Integer.parseInt(sb.toString());
	}
	
	private Long readLong() throws IOException {
		String num = new String();
		int sign = 1;
		read();
		if (ch == '-') {
			sign = -1;
			read();
		}
		while (Character.isDigit(ch)) {
			num += ch;
			read();
		}
		return new Long(num)*sign;
	}
	
	private Boolean readBoolean() throws IOException {
		boolean value = true;
		read();
		if (ch == 'f')
			value = false;
		read();
		return new Boolean(value);
	}
	
	private Float readFloat() throws IOException {
		StringBuilder sb = new StringBuilder();
		read();
		while (ch != 'e') {
			sb.append(ch);
			read();
		}
		return Float.parseFloat(sb.toString());
	}
	
	private Double readDouble() throws IOException {
		StringBuilder sb = new StringBuilder();
		read();
		while (ch != 'e') {
			sb.append(ch);
			read();
		}
		return Double.parseDouble(sb.toString());
	}

	private String readString() throws IOException {
		String slen = new String();
		while (Character.isDigit(ch)) {
			slen = slen + ch;
			read();
		}
		int len = new Integer(slen).intValue();
		String str = new String();
		for (int i = 0; i < len; i++) {
			str += read();
		}
		return str;
	}

	private Map<String, Object> readDictionary() throws IOException {
		HashMap<String,Object> dict = new HashMap<String,Object>();
		while (true) {
			Object key = readValue();
			if ((key == null) || !(key instanceof String))
				break;
			Object value = readValue();
			dict.put((String)key, value);
		}
		return dict;
	}
	
	@SuppressWarnings("unchecked")
	private ILambdaValue readLambda() throws IOException {
		String codepath = (String) readValue();
		List instructions = (List) readValue();
		Integer start = (Integer) readValue();
		Map<String,Object> dict = (Map<String, Object>) readValue();
		/* Eat final 'e' */
		read();
		return new LambdaValue(new VMCode(codepath, instructions), start, dict);
	}
	
	private IReferenceValue readReference() throws IOException {
		String name = (String) readValue();
		Integer index = (Integer) readValue();
		/* Eat final 'e' */
		read();
		return new ReferenceValue(name, index);
	}

	private Boolean readNil() throws IOException {
		read();
		return new Boolean(false);
	}

	public Object get() {
		return dataTree;
	}
	
	private Object readPointer() throws IOException {
		StringBuilder sb = new StringBuilder();
		read();
		while (ch != 'e') {
			sb.append(ch);
			read();
		}
		Integer index = Integer.parseInt(sb.toString());
		Object ret = directory.get(index);
		if (ret == null) throw new IOException("Unreferenced pointer to object "+index);
		return ret;
	}

}
