/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package serialization;

import interfaces.vm.util.ISerializer;
import interfaces.vm.valuetypes.ILambdaValue;
import interfaces.vm.valuetypes.IReferenceValue;

import java.io.OutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import com.fuzzbug.solenopsis.sdk.lib.IBaseSerializable;
import com.fuzzbug.solenopsis.sdk.lib.IConstants;
import com.fuzzbug.solenopsis.sdk.lib.IConstants.OPCODES;

import dot.codegenerator.Instruction;


public class DebugSerializer implements ISerializer {

	private PrintStream swr;
	
	private List<Object> directory;

	public DebugSerializer(OutputStream sprintr) {
		swr = new PrintStream(sprintr);
		directory = new ArrayList<Object>();
	}

	public DebugSerializer(PrintStream pprintr) {
		swr = pprintr;
		directory = new ArrayList<Object>();
	}

	@SuppressWarnings("unchecked")
	public void print(Object item) throws Exception {
		
		/* Simple types are serialized "by value" */
		
		if (item instanceof String) {
			printString((String)item);
			return;
		} else if (item instanceof Integer) {
			printInteger((Integer)item);
			return;
		} else if (item instanceof Long) {
			printLong((Long)item);
			return;
		} else if (item instanceof IConstants.OPCODES) {
			printOpcode((IConstants.OPCODES)item);
			return;
		} else if (item instanceof Float) {
			printFloat((Float)item);
			return;
		} else if (item instanceof Double) {
			printDouble((Double)item);
			return;
		} else if (item instanceof Boolean) {
			Boolean b = (Boolean) item;
			if(b == false) 
					printNil(null);
				else
					printBoolean((Boolean)item);
			return;
		} else if (item instanceof Instruction) {
			printInstruction((Instruction)item);
			return;
		} else if (item instanceof IReferenceValue) {
			printReference((IReferenceValue)item);
			return;
		} else if (item == null) {
			printNil(false);
			return;
		} else if (item instanceof IBaseSerializable) {
			print(((IBaseSerializable)item).toBaseValue());
			return;
		}
		
		/* Complex types can be referenced */
		for (int i=0; i < directory.size(); i++) {
			if (item == directory.get(i)) {
				//System.out.println(item+" of type"+item.getClass()+" is in cache at pos "+i+" total "+directory.size());
				printPointer(i);
				return;
			}
		} 
		
		directory.add(item);
		
		if (item instanceof Stack)
			printStack((Stack)item);
		else if (item instanceof List)
			printList((List)item);
		else if (item instanceof Map)
			printDictionary((Map)item);
		else if (item instanceof ILambdaValue)
			printLambda((ILambdaValue)item);
		else if (item == null)
			printNil(false);
		else {
			throw new Exception("(WARN) Don't know how to serialize "+item);
		}
	}

	private void printPointer(Integer index) throws Exception {
		swr.print("<pointer>");
		swr.print(index);
		swr.print("</pointer>");
	}

	private void printInstruction(Instruction item) throws Exception {
		swr.print("<instruction>");
		print(item.toBaseValue());
		swr.print("</instruction>");
	}

	private void printOpcode(OPCODES item) {
		swr.print("<opcode>");
		swr.print(item.ordinal());
		swr.print("</opcode>");
	}

	@SuppressWarnings("unchecked")
	private void printList(List item) throws Exception {
		swr.print("<list>");
		for (int i = 0; i < item.size(); i++)
			print(item.get(i));
		swr.print("</list>");
	}
	
	@SuppressWarnings("unchecked")
	private void printStack(Stack item) throws Exception {
		swr.print("<stack>");
		for (int i = 0; i < item.size(); i++)
			print(item.get(i));
		swr.print("</stack>");
	}

	@SuppressWarnings("unchecked")
	private void printDictionary(Map item) throws Exception {
		swr.print("</dictionary>");
		Iterator keys = item.keySet().iterator();
		while (keys.hasNext()) {
			Object key = keys.next();
			Object value = item.get(key);
			print(key);
			print(value);
		}
		swr.print("</dictionary>");
	}
	
	private void printLambda(ILambdaValue item) throws Exception {
		swr.print("<lambda>");
		print(item.getCode().getSourcePath());
		print(item.getCode().getInstructions());
		print(item.getLambdaStart());
		print(item.getBindings());
		swr.print("</lambda>");
	}
	
	private void printReference(IReferenceValue item) throws Exception {
		swr.print("<reference>");
		print(item.getReferencedName());
		print(item.getFrameIndex());
		swr.print("</reference>");
	}

	private void printString(String item) {
		String str = item.toString();
		swr.print(str.length());
		swr.print(':');
		swr.print(str);
	}

	private void printInteger(Integer item) {
		swr.print("<integer>");
		swr.print(item.toString());
		swr.print("</integer");
	}
	
	private void printLong(Long item) {
		swr.print("<long>");
		swr.print(item.toString());
		swr.print("</long>");
	}
	
	private void printFloat(Float item) {
		swr.print("<float>");
		swr.print(item.toString());
		swr.print("</float>");
	}
	
	private void printDouble(Double item) {
		swr.print("<double>");
		swr.print(item.toString());
		swr.print("</double");
	}
	
	private void printBoolean(Boolean item) {
		swr.print("<boolean>");
		if(item.booleanValue())
			swr.print("<true/>");
		else
			swr.print("</false>");
		swr.print("</boolean>");
	}

	private void printNil(Boolean item) {
		swr.print("<nil/>");
	}

}
