/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package serialization;

import interfaces.vm.util.ISerializer;
import interfaces.vm.valuetypes.ILambdaValue;
import interfaces.vm.valuetypes.IReferenceValue;

import java.io.OutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import com.fuzzbug.solenopsis.sdk.lib.IBaseSerializable;
import com.fuzzbug.solenopsis.sdk.lib.IConstants;
import com.fuzzbug.solenopsis.sdk.lib.IConstants.OPCODES;

import dot.codegenerator.Instruction;


public class Serializer implements ISerializer {

	private PrintStream swr;
	
	private List<Object> directory;

	public Serializer(OutputStream sprintr) {
		swr = new PrintStream(sprintr);
		directory = new ArrayList<Object>();
	}

	public Serializer(PrintStream pprintr) {
		swr = pprintr;
		directory = new ArrayList<Object>();
	}

	@SuppressWarnings("unchecked")
	public void print(Object item) throws Exception {
		
		/* Simple types are serialized "by value" */
		
		if (item instanceof String) {
			printString((String)item);
			return;
		} else if (item instanceof Integer) {
			printInteger((Integer)item);
			return;
		} else if (item instanceof Long) {
			printLong((Long)item);
			return;
		} else if (item instanceof IConstants.OPCODES) {
			printOpcode((IConstants.OPCODES)item);
			return;
		} else if (item instanceof Float) {
			printFloat((Float)item);
			return;
		} else if (item instanceof Double) {
			printDouble((Double)item);
			return;
		} else if (item instanceof Boolean) {
			Boolean b = (Boolean) item;
			if(b == false) 
					printNil(null);
				else
					printBoolean((Boolean)item);
			return;
		} else if (item instanceof Instruction) {
			printInstruction((Instruction)item);
			return;
		} else if (item instanceof IReferenceValue) {
			printReference((IReferenceValue)item);
			return;
		} else if (item == null) {
			printNil(false);
			return;
		}
		
		/* Complex types can be referenced */
		for (int i=0; i < directory.size(); i++) {
			if (item == directory.get(i)) {
				printPointer(i);
				return;
			}
		} 

		if (item instanceof Stack)
			printStack((Stack)item);
		else if (item instanceof List)
			printList((List)item);
		else if (item instanceof Map)
			printDictionary((Map)item);
		else if (item instanceof ILambdaValue)
			printLambda((ILambdaValue)item);
		else if (item instanceof IBaseSerializable) {
			item = ((IBaseSerializable)item).toBaseValue();
			print(item);
			return;
		} else if (item == null)
			printNil(false);
		else {
			throw new Exception("(WARN) Don't know how to serialize "+item);
		}
		
		directory.add(item);
	}

	private void printPointer(Integer index) throws Exception {
		swr.print('P');
		swr.print(index);
		swr.print('e');
	}

	private void printInstruction(Instruction item) throws Exception {
		swr.print('X');
		print(item.toBaseValue());
		swr.print('e');
	}

	private void printOpcode(OPCODES item) {
		swr.print('c');
		swr.print(item.ordinal());
		swr.print('e');
	}

	@SuppressWarnings("unchecked")
	private void printList(List item) throws Exception {
		swr.print('l');
		for (int i = 0; i < item.size(); i++)
			print(item.get(i));
		swr.print('e');
	}
	
	@SuppressWarnings("unchecked")
	private void printStack(Stack item) throws Exception {
		swr.print('s');
		for (int i = 0; i < item.size(); i++)
			print(item.get(i));
		swr.print('e');
	}

	@SuppressWarnings("unchecked")
	private void printDictionary(Map item) throws Exception {
		swr.print('d');
		Iterator keys = item.keySet().iterator();
		while (keys.hasNext()) {
			Object key = keys.next();
			Object value = item.get(key);
			print(key);
			print(value);
		}
		swr.print('e');
	}
	
	private void printLambda(ILambdaValue item) throws Exception {
		swr.print('x');
		print(item.getCode().getSourcePath());
		print(item.getCode().getInstructions());
		print(item.getLambdaStart());
		print(item.getBindings());
		swr.print('e');
	}
	
	private void printReference(IReferenceValue item) throws Exception {
		swr.print('r');
		print(item.getReferencedName());
		print(item.getFrameIndex());
		swr.print('e');
	}

	private void printString(String item) {
		String str = item.toString();
		swr.print(str.length());
		swr.print(':');
		swr.print(str);
	}

	private void printInteger(Integer item) {
		swr.print('i');
		swr.print(item.toString());
		swr.print('e');
	}
	
	private void printLong(Long item) {
		swr.print('I');
		swr.print(item.toString());
		swr.print('e');
	}
	
	private void printFloat(Float item) {
		swr.print('f');
		swr.print(item.toString());
		swr.print('e');
	}
	
	private void printDouble(Double item) {
		swr.print('F');
		swr.print(item.toString());
		swr.print('e');
	}
	
	private void printBoolean(Boolean item) {
		swr.print('o');
		if(item.booleanValue())
			swr.print("t");
		else
			swr.print("f");
		swr.print('e');
	}

	private void printNil(Boolean item) {
		swr.print('n');
		swr.print('e');
	}

}
