/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package shell;

import interfaces.exe.ICodeProvider;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import logging.Logger;
import providers.CodeProvider;
import providers.DefaultResourceProvider;
import providers.GenericResourceProvider;
import providers.VMFactory;
import services.Ants;
import services.BebopProxy;
import services.NodeDaemonProxy;
import services.StateSerialization;
import bebop.mailbox.MailService;

import com.fuzzbug.solenopsis.sdk.exe.INodeDaemon;
import com.fuzzbug.solenopsis.sdk.plugins.BasicPlugin;
import com.fuzzbug.solenopsis.sdk.plugins.IPlugin;
import com.fuzzbug.solenopsis.sdk.vm.IVM;

import core.signals.SignalEmitter;
import core.signals.signals.ExternCallSignalData;
import core.signals.signals.ExternFunctionCallSignalData;
import core.vm.lib.VMState;
import exe.NodeDaemon;
import exe.PlatformDaemon;
import exe.VMStats;

public class PlatformProxy extends BasicPlugin {
	
	private ICodeProvider cp;
	private MailService ms;
	private Hashtable<String,NodeDaemon> daemons;
	private PlatformDaemon sd;
	private IPlugin ss;
	private IPlugin ndp;
	private IPlugin ants;
	private Hashtable<String, IPlugin> sharedPlugins = new Hashtable<String, IPlugin>();
	
	public PlatformProxy() {
		super("platform", "Amos Brocco <amos.brocco@unifr.ch", "Bebop Platform Proxy", "1.0", false);
		registerMethods();
	}
	
	private Object getIfSet(Map<String, Object> data, String key, Object d) {
		if (data.containsKey(key))
			return data.get(key);
		else
			return d;
	}
	
	@SuppressWarnings("unchecked")
	public Object apply_init(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		try {
			String antcodepath = System.getProperty("user.dir");
			Integer port = 56333;
			String host = null;
			try {
				host = InetAddress.getLocalHost().getHostAddress();
			} catch (UnknownHostException e) {
				host = "localhost";
			}
			if ((args.size() == 1) && (args.get(0) instanceof Map)) {
				Map<String,Object> parameters = (Map<String,Object>) args.get(0);
				antcodepath = (String) getIfSet(parameters, "codepath", antcodepath);
				port = (Integer) getIfSet(parameters, "port", port);
				host = (String) getIfSet(parameters, "hostname", host);
				VMState.serializeFullState = getIfSet(parameters, "serializeFullState", false).equals(Boolean.FALSE) ? false : true;
			}
			DefaultResourceProvider.setProvider(new GenericResourceProvider(antcodepath));
			daemons = new Hashtable<String,NodeDaemon>();
			sd = new PlatformDaemon();
			cp = new CodeProvider();
			ms = MailService.start(host, port);
			ss = new StateSerialization();
			ndp = new NodeDaemonProxy(null);
			ants = new Ants(null);
		} catch (Exception e) {
			Logger.fatal("Failed to initialize platform; cannot continue.");
			e.printStackTrace();
			System.exit(1);
		}
		return 1;
	}
	
	public Object apply_getDaemons(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		List<String> did = new ArrayList<String>();
		did.addAll(daemons.keySet());
		return did;
	}
	
	public Object apply_addDaemon(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		synchronized (daemons) {
			String name = UUID.randomUUID().toString();
			if (args.size() == 1) {
				name = (String) args.get(0);
			}
			if (daemons.containsKey(name)) return daemons.get(name);
			NodeDaemon a = new NodeDaemon(name, cp);
			IPlugin bpa = new BebopProxy(cp, ms, a);
			a.registerPlugin(bpa);
			a.registerPlugin(ss);
			a.registerPlugin(ndp);
			a.registerPlugin(ants);
			daemons.put(a.getIdentifier(), a);
			sd.schedule(a);
			return new String(a.getIdentifier());
		}
	}
	
	public Object apply_removeDaemon(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		synchronized (daemons) {
			if (args.size() != 1) return false;
			String name = (String) args.get(0);
			if (!daemons.containsKey(name)) return false;
			NodeDaemon d = daemons.get(name);
			d.doFinalize();
			daemons.remove(d);
			sd.unschedule(d);
			return 1;
		}
	}
	
	@SuppressWarnings("unchecked")
	public Object apply_scheduleVM(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		NodeDaemon d = null;
		IVM nvm = null;
		synchronized (daemons) {
			if (args.size() < 2) return false;
			String dname = (String) args.get(0);
			if (!daemons.containsKey(dname)) return false;
			String cname = (String) args.get(1);
			d = daemons.get(dname);
			nvm = VMFactory.getVM(cp, cname, d);
			if (args.size() == 3) {
				Map<String,Object> gvar = (Map<String,Object>) args.get(2);
				for (String vname : gvar.keySet()) {
					nvm.getStack().setGValue(vname, gvar.get(vname));
				}
			}
		}
		d.schedule(nvm);
		return 1;
	}
	
	public Object apply_listPlugins(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 1) return false;
		String dname = (String) args.get(0);
		if (!daemons.containsKey(dname)) return false;
		NodeDaemon d = daemons.get(dname);
		List<String> pl = new ArrayList<String>();
		pl.addAll(d.getNamespaces());
		return pl;
	}
	
	public Object apply_listFunctions(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 2) return false;
		String dname = (String) args.get(0);
		if (!daemons.containsKey(dname)) return false;
		NodeDaemon d = daemons.get(dname);
		String namespace = (String) args.get(1);
		IPlugin pg = d.getPlugin(namespace);
		if (pg == null) return false;
		List<String> fl = new ArrayList<String>();
		for (String tag : pg.getTagsNames())
			fl.add(tag);
		return fl;
	}
	
	public Object apply_call(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() < 2) return false;
		String dname = (String) args.get(0);
		if (!daemons.containsKey(dname)) return false;
		NodeDaemon d = daemons.get(dname);
		String fname = (String) args.get(1);
		List<Object> cargs = null;
		if (args.size() > 2) {
			cargs = args.subList(2, args.size());
		} else {
			args = new ArrayList<Object>();
		}
		ExternCallSignalData data = new ExternFunctionCallSignalData(fname, cargs);
		return d.vmExternOpCall((SignalEmitter) vm, data);
	}
	
	public Object apply_registerPlugin(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 2) return false;
		String dname = (String) args.get(0);
		if (!daemons.containsKey(dname)) return false;
		NodeDaemon d = daemons.get(dname);
		String cName = (String) args.get(1);
		try {
			IPlugin p = (IPlugin) Class.forName(cName).getConstructor(INodeDaemon.class).newInstance(d);
			d.registerPlugin(p);
			return cName;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return false;
	}
	
	public Object apply_registerSharedPlugin(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		if (args.size() != 2) return false;
		String dname = (String) args.get(0);
		if (!daemons.containsKey(dname)) return false;
		NodeDaemon d = daemons.get(dname);
		String cName = (String) args.get(1);
		IPlugin p = null;
		if (sharedPlugins .containsKey(cName)) {
			p = sharedPlugins.get(cName);
		} else {
			p = (IPlugin) Class.forName(cName).getConstructor(INodeDaemon.class).newInstance(d);
			if (!p.isShareable()) return false;
			sharedPlugins.put(cName, p);
		}
		try {
			d.registerPlugin(p);
			return cName;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return false;
	}
	
	
	public Object apply_getStats(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		return VMStats.getStats();
	}
	
	public Object apply_run(INodeDaemon node, IVM vm,
			List<Object> args) throws Exception {
		sd.runSimulation();
		return 1;
	}
}
