/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package shell;


import java.io.ByteArrayInputStream;
import java.util.List;
import java.util.Scanner;

import logging.Logger;
import logging.StaticLogger;
import providers.GenericResourceProvider;
import providers.IResourceProvider;
import providers.VMFactory;
import cli.CLIOptions;
import cli.ICLIOptions;
import cli.IStringOption;
import cli.ISwitchOption;
import cli.InformationHeader;
import cli.StringOption;
import cli.SwitchOption;

import com.fuzzbug.solenopsis.sdk.lib.IInstruction;
import com.fuzzbug.solenopsis.sdk.vm.IVM;

import core.vm.lib.VMCode;
import core.vm.lib.VMStack;
import core.vm.lib.VMState;
import dot.compiler.Compiler;
import exe.NodeDaemon;

public class Shell {
	
	private NodeDaemon d;
	private VMStack stack;
	private IResourceProvider rp;
	private ICLIOptions cli;
	private ISwitchOption printcli;
	private ISwitchOption cliHelp;
	private ISwitchOption cliInfo;
	private IStringOption script;
	private ISwitchOption debug;
	
	
	public Shell(String[] args) {
		System.out.println("Solenopsis 2 Console - (c)2006-2009 Amos Brocco <amos.brocco@unifr.ch>");
		setupCLIParser(args);
		d = new NodeDaemon("<shell>", null);
		d.registerPlugin(new PlatformProxy());
		stack = new VMStack();
		rp = new GenericResourceProvider(System.getProperty("user.dir"));
		parseCLI();
	}
	
	private void setupCLIParser(String[] args) {
		cli = new CLIOptions(args);

		printcli = new SwitchOption(null, "printcli", "Outputs commandline");
		cli.addOption(printcli);
		
		debug = new SwitchOption(null, "debug", "Debug errors (print stack trace)");
		cli.addOption(debug);
		
		/* General options */
		cliHelp = new SwitchOption("h", "help", "Show this help");
		cli.addOption(cliHelp);
		cliInfo = new SwitchOption(null, "info",
				"Show copyright info and redistribution license");
		cli.addOption(cliInfo);
		script = new StringOption("s", "script", "Script to execute", null, false);
		cli.addOption(script);
	}
	
	protected void parseCLI() {
		if (!cli.parse() || cliHelp.getValue() || cliInfo.getValue()) {
			if (cliInfo.getValue())
				InformationHeader.print();
			else
				cli.printUsage();
			System.exit(1);
		}
		
		if (printcli.getValue())
			System.out.println(cli.toString());
		else if (script.getValue() != null) {
			System.out.format("Executing script '%s'\n", script.getValue());
			execute(String.format("#include \"%s\"", script.getValue()));
		}
		
	}

	public void cli() {
		for (;;) {
			Scanner scanner = new Scanner(System.in);
			System.out.print(">>> ");
			String command = scanner.nextLine();
			execute(command);
		}
	}
	
	private void execute(String command) {
		String xcommand = addVirtuals(command);
		ByteArrayInputStream bais = new ByteArrayInputStream(xcommand.getBytes());
		try {
			List<IInstruction> ilist = Compiler.compile(rp, "<shell>", bais);
			VMCode vc = new VMCode("<shell>", ilist);
			VMState st = new VMState(stack, vc, 0);
			IVM vm = VMFactory.getVM(st, d);
			d.schedule(vm);
			d.run();
			Object tos = stack.getTOS();
			tos = Boolean.FALSE.equals(tos) ? "nil" : tos;
			if (debug.getValue())
				System.out.println(tos);
		} catch (Exception e) {
			Logger.warning("Failed to execute command '%s', reason:\n%s", command, e.getMessage());
			if (debug.getValue())
				e.printStackTrace();
		}
	}
	
	private String addVirtuals(String command) {
		StringBuilder b = new StringBuilder();
		for (String v : stack.getGVariables()) {
			b.append(String.format("(virtual %s)\n", v));
		}
		b.append(String.format("%s", command));
		return b.toString();
	}

	public static void main(String[] args) throws Exception {
		StaticLogger.addOutput(System.out);
		StaticLogger.setDefaultLevel(0);
		Shell s = new Shell(args);
		s.cli();
	}

}
