/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package com.fuzzbug.solenopsis.sdk.vm;

import interfaces.vm.valuetypes.ILambdaValue;
import interfaces.vm.valuetypes.IReferenceValue;

import java.util.List;
import java.util.Set;

import com.fuzzbug.solenopsis.sdk.lib.IBaseSerializable;



public interface IVMStack extends IBaseSerializable {

	/**
	 * Returns the value of a variable from the global environment
	 * @param name	The name of the variable
	 * @return			The value of the variable
	 */
	public Object getGValue(String name);

	/**
	 * Returns the lambda object associated with the current stack frame
	 * @return	The lambda object
	 */
	public ILambdaValue getLambda();
	
	/**
	 * Returns true if the current stack frame is local (generated by LOCAL)
	 * @return
	 */
	public boolean isLocalStackFrame();

	/**
	 * Returns the current value of the reference value given
	 * @param ref		Reference value object
	 * @return			The value or null if it's not defined
	 */
	public Object getReferenceValue(IReferenceValue ref);

	/**
	 * Returns the return program counter of the current frame
	 * @return	The program counter value
	 */
	public int getReturnPC();
	
	/**
	 * Returns the index of the top frame
	 * @return	The frame index
	 */
	public Integer getTopFrameIndex();
	
	/**
	 * Returns the value on top of the current active stack
	 * @return		The value on top of the stack
	 */
	public Object getTOS();

	/**
	 * Returns the value of a variable (recursively searching in the 
	 * environments)
	 * @param value	The name of the variable
	 * @return			The value of the variable or null if it's not defined
	 */
	public Object getValue(String value);

	public boolean isEmpty();

	/**
	 * Pops i values from the active stack
	 * @param i	Number of objects to be popped
	 * @return		A list of popped values
	 */
	public List<Object> popNValues(int i);

	/**
	 * Pops the current stack frame
	 * @return	The stack frame popped
	 */
	public IVMStackFrame popStackFrame();

	/**
	 * Pops a value from the current active stack
	 * @return	The popped value
	 */
	public Object popValue();

	/**
	 * Pushes multiple values on the stack
	 * @param fargs	A list of the values to be pushed
	 */
	public void pushNValues(List<Object> fargs);

	/**
	 * Creates a new stackframe, with return pc
	 * @param pc	The return program counter
	 * @return		The newly created stack frame
	 */
	public IVMStackFrame pushLocalStackFrame();	
	
	/**
	 * Creates a new stackframe, with return pc
	 * @param pc	The return program counter
	 * @return		The newly created stack frame
	 */
	public IVMStackFrame pushStackFrame(int pc);

	/**
	 * Creates a new stackframe, with return pc and associated lambda
	 * @param pc	The return program counter
	 * @param lambda	The associated lambda object
	 * @return		The newly created stack frame
	 */
	public IVMStackFrame pushStackFrame(int pc, ILambdaValue lambda);
	
	/**
	 * Pushes a value on the current active stack
	 * @param argument	The value to be pushed
	 */
	public void pushValue(Object argument);

	/**
	 * Sets the value of a variable in the global environment
	 * @param value	The name of the variable
	 * @param value2	The value of the variable
	 */
	public void setGValue(String name, Object value);

	/**
	 * Sets the value of a referenced variable
	 * @param name		The name of the variable
	 * @param value	The value of the variable
	 */
	public void setReferenceValue(String name, Object value);

	/**
	 * Sets the value of a variable
	 * @param value	The name of the variable
	 * @param value2	The value of the variable
	 */
	public void setValue(String name, Object value);
	
	/**
	 * Defines a value in the main stack frame (global environment)
	 * @param name		The name of the variable
	 * @param value	The value of the variable
	 */
	public void storeGValue(String name, Object value);
	
	/**
	 * Stores a value in the current stack frame
	 * @param name		The name of the variable
	 * @param value	The value of the variable
	 */
	public void storeValue(String name, Object value); 
	
	
	public Set<String> getGVariables();

	public void updateAllLambdaBindings(); 

}
